class CurvemapView
{
  int _width, _height;
  int _tab_width, _tab_height, _tab_offset;
  int[] _frame_pos, _curvemap_pos, _curvemap_pos_selected_group, _curvemap_pos_computation_group;
  int _frame_width, _frame_height;
  boolean _render_comp_group_frame;
  int _curvemap_width, _curvemap_height;
  int[] _upper_pane_pos, _lower_pane_pos;
  int _upper_pane_width, _upper_pane_height, _lower_pane_width, _lower_pane_height;

  int _filled_plot_timestep_width, _filled_plot_width, _filled_plot_height;
  int _filled_plot_offset;
  int _overlay_plot_timestep_width, _overlay_plot_width, _overlay_plot_height;
  int _overlay_plot_offset;
  int _group_cells_rendered_counter;

  boolean _cell_rolled_over, _coi_selected;
  int _rolled_over_cell, _rolled_over_cell_index;
  String _rolled_over_cell_embryo_id;

  int _comp_group_map_width;
  int[] _comp_group_map_pos;
  ComputationGroupEmbryoView _comp_group_map;

  int _save_load_view_width;
  int[] _save_load_view_pos;
  SaveLoadView _save_load_view;

  int[] _delete_column_pos;
  int _delete_column_width, _num_flagged_for_removal;

  int _scroll_tick_width;
  int _num_to_render_x;
  int _start_render_index_x;
  int _scroll_tick_interval_x, _scroll_length_x;
  int[] _scroll_pos_x;
  boolean _over_x_scroll;
  int _over_x_scroll_pos;
  int _num_to_render_y;
  int _start_render_index_y;
  int _scroll_tick_interval_y, _scroll_length_y;
  int[] _scroll_pos_y;
  boolean _over_y_scroll;
  int _over_y_scroll_pos;
  boolean _first_time_through;

  int _seperator_line_pos;


  CurvemapView()
  {
    _tab_width = CURVEMAP_VIEW_TAB_WIDTH;
    _tab_height = CURVEMAP_VIEW_TAB_HEIGHT;
    _tab_offset = CURVEMAP_VIEW_TAB_OFFSET;

    _frame_pos = new int[2];
    _frame_pos[0] = 0;
    _frame_pos[1] = _tab_height;

    _seperator_line_pos = 195;

    _upper_pane_pos = new int[2];
    _lower_pane_pos = new int[2];

    _created_group = new Group( "new" );

    _num_flagged_for_removal = 0;
    _group_cells_rendered_counter = 0;

    _cell_rolled_over = false;
    _rolled_over_cell = -1;
    _render_comp_group_frame = true;

    _num_to_render_x = _start_render_index_x = _num_to_render_y = _start_render_index_y = 0;

    _scroll_tick_width = SCROLL_WIDTH;
    _scroll_tick_interval_x = _scroll_length_x = _scroll_tick_interval_y = _scroll_length_y = 0;

    _over_x_scroll = _over_y_scroll = false;
    _over_x_scroll_pos = _over_y_scroll_pos = 0;

    _first_time_through = true;
    _coi_selected = false;

    _scroll_pos_x = new int[2];
    _scroll_pos_y = new int[2];

    _delete_column_pos = new int[2];
    _delete_column_pos[0] = CURVEMAP_VIEW_PANE_BORDER;
    _delete_column_pos[1] = CURVEMAP_VIEW_PANE_BORDER;
    _delete_column_width = CURVEMAP_VIEW_DELETE_COLUMN_WIDTH;

    _curvemap_pos = new int[2];
    _curvemap_pos[0] = _delete_column_pos[0] + _delete_column_width + CURVEMAP_VIEW_PANE_BORDER + CURVEMAP_VIEW_INSET/2;
    _curvemap_pos[1] = _delete_column_pos[1] + CURVEMAP_VIEW_SCROLL_OFFSET + SCROLL_WIDTH;

    _comp_group_map = new ComputationGroupEmbryoView();
    _comp_group_map_pos = new int[2];
    _comp_group_map_pos[0] = CURVEMAP_VIEW_PANE_BORDER;
    _comp_group_map_pos[1] = CURVEMAP_VIEW_PANE_BORDER;    

    _save_load_view = new SaveLoadView();
    _save_load_view_pos = new int[2];
    _save_load_view_pos[0] = CURVEMAP_VIEW_PANE_BORDER;
    _save_load_view_pos[1] = CURVEMAP_VIEW_PANE_BORDER;    
  }

  boolean cellRolledOver()
  {
    return _cell_rolled_over;
  }

  int rolledOverCellIndex()
  {
    return _rolled_over_cell_index;
  }

  String rolledOverCellEmbryoID()
  {
    return _rolled_over_cell_embryo_id;
  }

  boolean selectedCellRolledOver()
  {
    return ( _cell_rolled_over && (_rolled_over_cell == -1) );
  }

  boolean renderCompGroupFrame()
  {
    return _render_comp_group_frame;
  }

  void setDimensions( int w, int h )
  {
    _width = w;
    _height = h;

    _frame_width = _width;
    _frame_height = _height - _tab_height;

    _upper_pane_pos[0] = 0;
    _upper_pane_pos[1] = 0;

    _lower_pane_pos[0] = 0;
    _lower_pane_pos[1] = _seperator_line_pos+5;

    _upper_pane_width = _frame_width;
    _lower_pane_width = _frame_width;
    _upper_pane_height = _lower_pane_pos[1] - _upper_pane_pos[1];
    _lower_pane_height = _frame_height - _upper_pane_height;

    _curvemap_width = _lower_pane_width - _curvemap_pos[0]-CURVEMAP_VIEW_PANE_BORDER /* - CURVEMAP_VIEW_SCROLL_OFFSET - SCROLL_WIDTH*/;
    _curvemap_height = _lower_pane_height - _curvemap_pos[1] - CURVEMAP_VIEW_PANE_BORDER /*CURVEMAP_VIEW_SCROLL_OFFSET - SCROLL_WIDTH*/;

    _comp_group_map.setDimensions( _upper_pane_width-2*CURVEMAP_VIEW_PANE_BORDER, _upper_pane_height-2*CURVEMAP_VIEW_PANE_BORDER );
    _comp_group_map.setTextOffset( _curvemap_view_origin[0]+_frame_pos[0]+_save_load_view_pos[0], _curvemap_view_origin[1]+_frame_pos[1]+_save_load_view_pos[1] );

    _save_load_view.setDimensions( _upper_pane_width-2*CURVEMAP_VIEW_PANE_BORDER, _upper_pane_height-2*CURVEMAP_VIEW_PANE_BORDER );
    _save_load_view.setTextOffset( _curvemap_view_origin[0]+_frame_pos[0]+_save_load_view_pos[0], _curvemap_view_origin[1]+_frame_pos[1]+_save_load_view_pos[1] );

    if ( !_first_time_through )
      initializeCompGroup();
    setScrollDimensions();   
  }

  void initializeCompGroup()
  {
    Aggregation a = _aggregations[_selected_aggregation];
    GroupElement cell_of_interest_ge =  a.getCellOfInterest();
    _comp_group_map.initialize( a.getComputationGroup(a.cellOfInterestIndex()), cell_of_interest_ge );    
  }

  void initializations()
  {
    _filled_plot_timestep_width = floor((float)CURVEMAP_VIEW_FILLED_TIMESERIES_WIDTH/(float)(NUM_TIMESTEPS-1));
    _filled_plot_width = _filled_plot_timestep_width * (NUM_TIMESTEPS-1);
    _filled_plot_height = _filled_plot_width;
    _filled_plot_offset = CURVEMAP_VIEW_PLOT_OFFSET;

    _overlay_plot_timestep_width = _filled_plot_timestep_width;
    _overlay_plot_width = _filled_plot_width;
    _overlay_plot_height = _filled_plot_height;
    _overlay_plot_offset = _filled_plot_offset;

    initializeCompGroup();
  }

  void setScrollDimensions()
  {    
    // compute the number of genes that can be shown
    _num_to_render_x = min( floor((float)_curvemap_width / (float)(_filled_plot_width+_filled_plot_offset)), NUM_GENES );
    _start_render_index_x = min( _start_render_index_x, (NUM_GENES-_num_to_render_x) );

    _scroll_length_x = (_filled_plot_width + _filled_plot_offset)*_num_to_render_x - _filled_plot_offset;
    _scroll_tick_interval_x = round( (float)_scroll_length_x / (float)NUM_GENES );
    _scroll_length_x = _scroll_tick_interval_x * NUM_GENES;
    _scroll_pos_x[0] = _curvemap_pos[0];
    _scroll_pos_x[1] = _curvemap_pos[1]-CURVEMAP_VIEW_SCROLL_OFFSET-SCROLL_WIDTH;

    setScrollYDimensions();
  }

  void setScrollYDimensions()
  {
    if ( _first_time_through ) _first_time_through = false;
    else countGroupCellsRendered();

    // compute the number of cells that can be shown
    int cm_cells_h = _curvemap_height - 2*_filled_plot_height - CURVEMAP_VIEW_COI_BORDER - CURVEMAP_VIEW_OVERLAY_BORDER;
    _num_to_render_y = min( floor((float)cm_cells_h  / (float)(_filled_plot_height+_filled_plot_offset)), _group_cells_rendered_counter );
    _start_render_index_y = min( _start_render_index_y, (_group_cells_rendered_counter-_num_to_render_y) );

    _scroll_length_y = (_filled_plot_width + _filled_plot_offset)*_num_to_render_y - _filled_plot_offset;
    _scroll_tick_interval_y = round( (float)_scroll_length_y / (float)_group_cells_rendered_counter );
    _scroll_length_y = _scroll_tick_interval_y * _group_cells_rendered_counter;
    _scroll_pos_y[0] = _curvemap_pos[0] + (_filled_plot_width+_filled_plot_offset)*_num_to_render_x - _filled_plot_offset + CURVEMAP_VIEW_SCROLL_OFFSET;
    _scroll_pos_y[1] = _curvemap_pos[1] + _filled_plot_width + CURVEMAP_VIEW_COI_BORDER;
  }

  void countGroupCellsRendered()
  {
    Group g;
    if ( _render_comp_group_frame )
    {
      Aggregation a = _aggregations[_selected_aggregation];
      g = a.getComputationGroup(a.cellOfInterestIndex());
    }
    else 
      g = _created_group;

    _group_cells_rendered_counter = 0;
    for ( int i = 0; i < g.getNumElements(); i++ )
      if ( g.getElement(i).render() ) ++_group_cells_rendered_counter;
  }

  void render()
  {
    noSmooth();
    int x, y, offset, tx_offset=0, ty_offset=0;

    //
    // render the frame
    //
    if( _use_text_offset )
    {
      tx_offset = _curvemap_view_origin[0] + _frame_pos[0];
      ty_offset = _curvemap_view_origin[1] + _frame_pos[1];
    }
    int y_bottom_offset = 0;
    if ( !_use_text_offset )
    { 
      ty_offset = -PDF_OFFSET;
      y_bottom_offset = PDF_OFFSET-1;
    }

    pushMatrix();    
    translate( _frame_pos[0], _frame_pos[1] );

    stroke( CURVEMAP_VIEW_FRAME_COLOR );
    strokeWeight( CURVEMAP_VIEW_FRAME_WEIGHT );
    noFill();
    rectMode( CORNER );
    rect( 0, 0, _frame_width, _frame_height );

    //
    // render the title
    //
    x = _frame_width;
    y = 0;
    textFont( _verdana_font_16 );
    fill( TITLE_COLOR );
    textAlign( RIGHT, BOTTOM );
    text( "Curvemap", x+tx_offset, y+ty_offset+y_bottom_offset );
    //strokeWeight( 1 );
    //stroke(  KEY_COLOR );
    //line( x, y, x+textWidth("Unrolled View"), y );

    //
    // render the tabs
    //
    textAlign( CENTER, CENTER );
    textFont(_pixel_font_8);

    fill( BACKGROUND_COLOR );
    rect( 0, 0, _tab_width, -_tab_height );     
    fill( CURVEMAP_VIEW_TEXT_COLOR );
    text( "AGGREGATION GROUP", tx_offset + _tab_width/2, ty_offset - _tab_height/2 );

    fill( BACKGROUND_COLOR );
    rect( _tab_width+_tab_offset, 0, _tab_width, -_tab_height );
    fill( CURVEMAP_VIEW_TEXT_COLOR );
    text( "CREATED GROUP", tx_offset + _tab_width/2 + _tab_width+_tab_offset, ty_offset - _tab_height/2 );

    //
    // showcase the selected tab
    //
    stroke( BACKGROUND_COLOR );
    strokeWeight( 4 );
    if ( _render_comp_group_frame ) line( 1, 0, _tab_width, 0 );
    else line( _tab_width+_tab_offset+1, 0, 2*_tab_width+_tab_offset, 0 );



    //
    // render the line seperating the frames
    //
    strokeWeight( CURVEMAP_VIEW_FRAME_WEIGHT );
    stroke( CURVEMAP_VIEW_FRAME_COLOR );
    line( 0, _seperator_line_pos, _frame_width, _seperator_line_pos );

    //
    // render the lower frame
    pushMatrix();
    translate( _lower_pane_pos[0], _lower_pane_pos[1] );

    //
    // render the curvemap
    //
    if ( _render_comp_group_frame ) renderComputationGroupFrame();
    else renderSelectedGroupFrame();

    //
    // done with the lower frame
    //
    popMatrix();

    //
    // render the upper pane
    //
    pushMatrix();
    translate( _upper_pane_pos[0], _upper_pane_pos[1] );

    // render the aggregation group
    if ( _render_comp_group_frame ) 
    {
      pushMatrix();
      translate( _comp_group_map_pos[0], _comp_group_map_pos[1] );

      Aggregation a = _aggregations[_selected_aggregation];
      GroupElement cell_of_interest_ge =  a.getCellOfInterest();
      _comp_group_map.render( _cell_rolled_over, _rolled_over_cell );

      popMatrix();
    }

    // render the save load view
    else
    {
      pushMatrix();
      translate( _save_load_view_pos[0], _save_load_view_pos[1] );

      _save_load_view.render();

      popMatrix();
    }

    //
    // done with the upper frame
    //
    popMatrix();

    //
    // done rendering the frame
    //
    popMatrix();
  }

  void renderComputationGroupFrame()
  {
    Aggregation a = _aggregations[_selected_aggregation];
    GroupElement cell_of_interest_ge =  a.getCellOfInterest();

    //
    // render the delete column
    //
    renderDeleteColumn( _delete_column_pos[0]+_delete_column_width/2, _delete_column_pos[1], a.getComputationGroup(a.cellOfInterestIndex()) );

    // 
    // render the curvemap
    //
    Embryo e = _embryos[a.getSelectedEmbryo()];
    Cell cell_of_interest = e.getCell(e.cellID2Index(cell_of_interest_ge.getCellID()));
    renderCurvemap( cell_of_interest, a.getComputationGroup(a.cellOfInterestIndex()), _curvemap_pos[0], _curvemap_pos[1] );
  }

  void renderSelectedGroupFrame()
  {
    Aggregation a = _aggregations[_selected_aggregation];
    GroupElement cell_of_interest_ge =  a.getCellOfInterest();

    //
    // render the delete column
    //
    renderDeleteColumn( _delete_column_pos[0]+_delete_column_width/2, _delete_column_pos[1], _created_group );

    Embryo e = _embryos[a.getSelectedEmbryo()];   
    Cell cell_of_interest = e.getCell(e.cellID2Index(cell_of_interest_ge.getCellID()));
    renderCurvemap( cell_of_interest, _created_group, _curvemap_pos[0], _curvemap_pos[1] );
  }

  void renderDeleteColumn( int x, int y, Group group )
  {   
    if ( _num_to_render_y == 0 ) return;

    y += 3*_filled_plot_height/2 + CURVEMAP_VIEW_COI_BORDER + CURVEMAP_VIEW_SCROLL_OFFSET + SCROLL_WIDTH;

    GroupElement ge;
    int i = -1;
    int counter = -1;
    _num_flagged_for_removal = 0;
    while ( counter < _start_render_index_y )
    {
      ++i;
      ge = group.getElement(i);
      if ( ge.render() )
      {
        ++counter;
        if ( ge.flaggedForRemoval() ) ++_num_flagged_for_removal;
      }
    }

    while ( counter < _start_render_index_y+_num_to_render_y )
    {
      ge = group.getElement(i);
      ++i;
      if ( !ge.render() ) continue;

      noFill();
      stroke( CURVEMAP_VIEW_DELETE_OUTLINE_COLOR );
      strokeWeight( CURVEMAP_VIEW_DELETE_OUTLINE_WEIGHT );

      ellipse( x, y, CURVEMAP_VIEW_DELETE_OUTLINE_WIDTH, CURVEMAP_VIEW_DELETE_OUTLINE_WIDTH );

      if ( ge.flaggedForRemoval() )
      {
        stroke( CURVEMAP_VIEW_DELETE_OUTLINE_COLOR );
        fill( CURVEMAP_VIEW_DELETE_FILL_COLOR );

        ellipse( x, y, CURVEMAP_VIEW_DELETE_FILL_WIDTH, CURVEMAP_VIEW_DELETE_FILL_WIDTH );
        ++_num_flagged_for_removal;
      }

      y += _filled_plot_height + CURVEMAP_VIEW_GROUP_CELL_BORDER;
      ++counter;
    }

    //
    // render the button
    //
    int tx_offset=0, ty_offset=0;    
    if( _use_text_offset )
    {
      tx_offset = _curvemap_view_origin[0] + _frame_pos[0] + _lower_pane_pos[0];
      ty_offset = _curvemap_view_origin[1] + _frame_pos[1] + _lower_pane_pos[1];
    }
    int y_bottom_offset = 0;
    if ( !_use_text_offset )
    { 
      ty_offset = -PDF_OFFSET;
      y_bottom_offset = PDF_OFFSET-1;
    }

    textAlign( CENTER, CENTER );
    textFont( _pixel_font_8 );
    stroke( CURVEMAP_VIEW_DELETE_OUTLINE_COLOR );
    strokeWeight( CURVEMAP_VIEW_DELETE_OUTLINE_WEIGHT );

    y -= (_filled_plot_height/3+CURVEMAP_VIEW_GROUP_CELL_BORDER);    
    fill( CURVEMAP_VIEW_DELETE_FILL_COLOR );
    rect( x-CURVEMAP_VIEW_DELETE_BUTTON_WIDTH/2+1, y, CURVEMAP_VIEW_DELETE_BUTTON_WIDTH, CURVEMAP_VIEW_DELETE_BUTTON_HEIGHT );    
    if ( _num_to_render_y != 0 ) fill( BACKGROUND_COLOR );
    else fill( CURVEMAP_VIEW_DELETE_OUTLINE_COLOR );
    text( "all", x+tx_offset+2, y+ty_offset+CURVEMAP_VIEW_DELETE_BUTTON_HEIGHT/2 );

    y += 3*CURVEMAP_VIEW_DELETE_BUTTON_HEIGHT/2;    
    fill( CURVEMAP_VIEW_DELETE_FILL_COLOR );
    rect( x-CURVEMAP_VIEW_DELETE_BUTTON_WIDTH/2+1, y, CURVEMAP_VIEW_DELETE_BUTTON_WIDTH, CURVEMAP_VIEW_DELETE_BUTTON_HEIGHT );    
    if ( _num_flagged_for_removal != 0 ) fill( BACKGROUND_COLOR );
    else fill( CURVEMAP_VIEW_DELETE_OUTLINE_COLOR );
    text( "remove", x+tx_offset+2, y+ty_offset+CURVEMAP_VIEW_DELETE_BUTTON_HEIGHT/2 );
  }

  void renderCurvemap( Cell cell_of_interest, Group group, int x, int y )
  {
    pushMatrix();
    translate( x, y );

    //
    // render the cell of interest profile
    //    
    x = 0;
    y = 0;
    renderCellsTimeSeries( x, y, cell_of_interest, true, -1, -1, null );    
    y += _filled_plot_height + CURVEMAP_VIEW_COI_BORDER;

    //
    // render the group profiles
    //
    if ( _num_to_render_y != 0 )
    {
      GroupElement ge;
      int i = -1;
      int counter = -1;
      while ( counter < _start_render_index_y )
      {
        ++i;
        ge = group.getElement(i);
        if ( ge.render() ) ++counter;
      }

      while ( counter < _start_render_index_y+_num_to_render_y )
      {
        ge = group.getElement(i);
        ++i;
        if ( !ge.render() ) continue;

        Embryo e = _embryos[embryoID2Index(ge.getEmbryoID())];
        renderCellsTimeSeries( x, y, e.getCell(e.cellID2Index(ge.getCellID())), false, counter, i-1, ge );
        y += _filled_plot_height + CURVEMAP_VIEW_GROUP_CELL_BORDER;
        ++counter;
      }
    }

    //
    // render the overlay plots
    //
    y += ( CURVEMAP_VIEW_OVERLAY_BORDER - CURVEMAP_VIEW_GROUP_CELL_BORDER );
    renderOverlayPlots( cell_of_interest, group, x, y );

    popMatrix();

    //
    // render the scroll bars
    //
    renderCurvemapScrollBars();
  }

  void renderCellsTimeSeries( int x, int y, Cell c, boolean coi, int group_counter, int indi_counter, GroupElement ge )
  {
    int tx_offset=0, ty_offset=0;
    if( _use_text_offset )
    {
      tx_offset = _curvemap_view_origin[0] + _frame_pos[0] + _lower_pane_pos[0] + _curvemap_pos[0];
      ty_offset = _curvemap_view_origin[1] + _frame_pos[1] + _lower_pane_pos[1] + _curvemap_pos[1];
    }
    int y_bottom_offset = 0;
    if ( !_use_text_offset )
    { 
      ty_offset = -PDF_OFFSET;
      y_bottom_offset = PDF_OFFSET-1;
    }

    //
    // render the gene names if this is the COI
    //
    if ( coi )
    {
      int tx = tx_offset + x + _filled_plot_width/2;
      int ty = ty_offset + y - 2;

      textFont( _pixel_font_8 );
      fill( CURVEMAP_VIEW_TEXT_COLOR );
      textAlign( CENTER, BOTTOM );
      for ( int i = _start_render_index_x; i < _start_render_index_x+_num_to_render_x; i++ )
      {
        text( _gene_names[i], tx, ty+y_bottom_offset );
        tx += _filled_plot_width + _filled_plot_offset;
      }
    }

    // render the cell id name
    Aggregation a = _aggregations[_selected_aggregation];
    boolean cg_cell_rolled, loaded_group_rolled;
    textFont(_pixel_font_8);
    if ( coi )
    { 
      cg_cell_rolled = (!_render_comp_group_frame && _embryo_view.cellRolledOver() && 
        (_embryo_view.rolledOverCell() == _embryos[a.getSelectedEmbryo()].cellID2Index(c.getID())) && (c.getEmbryoID().equals(_embryos[a.getSelectedEmbryo()].getID())) );
      fill( CURVEMAP_VIEW_SELECTED_COLOR );
      if ( (_cell_rolled_over && (_rolled_over_cell == -1)) || _coi_selected || _comp_group_map.rolledOverCOI() || cg_cell_rolled )
        textFont(_pixel_font_8b);
    }
    else
    {
      cg_cell_rolled = (!_render_comp_group_frame && _embryo_view.cellRolledOverInCreatedGroup() && 
        (_embryo_view.rolledOverCell() == _embryos[a.getSelectedEmbryo()].cellID2Index(c.getID())) && (c.getEmbryoID().equals(_embryos[a.getSelectedEmbryo()].getID())) );
      loaded_group_rolled = _save_load_view.loadedGroupRolledOver() && (_save_load_view.rolledOverLoadedGroup() == ge.existingGroupIndex());
      fill( CURVEMAP_VIEW_TEXT_COLOR );
      if ( (_cell_rolled_over && (_rolled_over_cell == group_counter)) || (_comp_group_map.rolledOverGroupCell() && _comp_group_map.groupCellRolledOver() == indi_counter) || cg_cell_rolled || loaded_group_rolled )
      {
        fill( CURVEMAP_VIEW_OVERLAY_PLOT_ROLLED_OVERLINE_COLOR );
        textFont(_pixel_font_8b);
      }
    }

    int l_y = ty_offset+y+2;
    textAlign( RIGHT, TOP );
    text( c.getEmbryoID(), tx_offset+x-2, l_y );
    l_y += 10;
    textAlign( RIGHT, TOP );
    text( c.getID(), tx_offset+x-2, l_y );
    if ( !coi && ge.fromAnExistingGroup() )
    {
      l_y += 10;
      Group g = (Group)_existing_groups.get(ge.existingGroupIndex());
      textAlign( RIGHT, TOP );
      fill( g.getColor() );
      text( g.getID(), tx_offset+x-2, l_y );
    }

    int px = x;
    boolean in_group;
    color col;
    for ( int i = _start_render_index_x; i < _start_render_index_x+_num_to_render_x; i++ )
    {
      if ( !coi ) in_group = ge.fromAnExistingGroup();
      else in_group = false;
      if ( in_group ) col = ((Group)_existing_groups.get(ge.existingGroupIndex())).getColor();
      else col = color(0);
        
      renderSingleFilledTimeSeries( px, y, c, i, in_group, col );
      px += _filled_plot_offset + _filled_plot_width; 
    }  
  }

  void renderSingleFilledTimeSeries( int x, int y, Cell c, int gene, boolean in_group, color col )
  {
    //
    // get the min and max values for this plot
    //    
    float[] min_max_v = c.getTimeSieresMinMax(gene);

    //
    // render the background color
    //
    fill( CURVEMAP_VIEW_PLOT_BACKGROUD_COLOR  );
    noStroke();
    rect( x, y, _filled_plot_width, _filled_plot_height );

    //
    // render the plot
    //
    color fill_color;
    if ( in_group ) fill_color = lerpColor( col, CURVEMAP_VIEW_PLOT_FILL_COLOR, 0.75 );
    else fill_color = CURVEMAP_VIEW_PLOT_FILL_COLOR;
    
    fill( CURVEMAP_VIEW_PLOT_FILL_COLOR  );
    float v, next_v;
    float v_diff = min_max_v[1] - min_max_v[0];
    int vert_y, next_vert_y;
    int vert_x = x;
    for ( int i = 0; i < NUM_TIMESTEPS-1; i++ )
    {
      v = c.getProfileValue( gene, i );
      next_v = c.getProfileValue( gene, i+1 );

      // check if we have missing data for this interval
      if ( Float.isNaN(v) || Float.isNaN(next_v) ) continue;

      // render the interval fill
      fill( fill_color  );
      noStroke();

      beginShape(); 

      //vert_y = y + round( (1.0 - (v - min_max_v[0]) / v_diff) * (float)_filled_plot_height);
      //next_vert_y = y + round( (1.0 - (next_v - min_max_v[0]) / v_diff) * (float)_filled_plot_height);

      vert_y = y + round( (1.0 - min(v, 1.0)) * (float)_filled_plot_height);
      next_vert_y = y + round( (1.0 - min(next_v, 1.0)) * (float)_filled_plot_height);

      vertex( vert_x, y+_filled_plot_height );
      vertex( vert_x+_filled_plot_timestep_width, y+_filled_plot_height );
      vertex( vert_x+_filled_plot_timestep_width, next_vert_y );
      vertex( vert_x, vert_y );

      endShape( CLOSE );

      // render the interval line
      noFill();
      stroke( CURVEMAP_VIEW_PLOT_LINE_COLOR  );
      strokeWeight( CURVEMAP_VIEW_PLOT_LINE_WEIGHT );
      line( vert_x, vert_y, vert_x+_filled_plot_timestep_width, next_vert_y );

      vert_x += _filled_plot_timestep_width; 
    }

    //
    // outline the plot
    //
    noFill();
    stroke( CURVEMAP_VIEW_PLOT_OUTLINE_COLOR );
    strokeWeight( CURVEMAP_VIEW_PLOT_OUTLINE_WEIGHT );
    rect( x, y, _filled_plot_width, _filled_plot_height );
  }

  void renderOverlayPlots( Cell cell_of_interest, Group group, int x, int y )
  {
    int tx_offset=0, ty_offset=0;
    if( _use_text_offset )
    {
      tx_offset = _curvemap_view_origin[0] + _frame_pos[0] + _lower_pane_pos[0] + _curvemap_pos[0];
      ty_offset = _curvemap_view_origin[1] + _frame_pos[1] + _lower_pane_pos[1] + _curvemap_pos[1];
    }
    int y_bottom_offset = 0;
    if ( !_use_text_offset )
    { 
      ty_offset = -PDF_OFFSET;
      y_bottom_offset = PDF_OFFSET-1;
    }

    //
    // render the gene names 
    //
    if ( _num_to_render_y != 0 )
    {
      int tx = tx_offset + x + _filled_plot_width/2;
      int ty = ty_offset + y - 12;

      textFont( _pixel_font_8 );
      fill( CURVEMAP_VIEW_TEXT_COLOR );
      textAlign( CENTER, BOTTOM );
      for ( int i = _start_render_index_x; i < _start_render_index_x+_num_to_render_x; i++ )
      {
        text( _gene_names[i], tx, ty+y_bottom_offset );
        tx += _filled_plot_width + _filled_plot_offset;
      }
    }

    float[] min_max_v = new float[2];
    float[] tmp_min_max_v;
    GroupElement ge;
    Aggregation a = _aggregations[_selected_aggregation];
    int px = x;
    for ( int g = _start_render_index_x; g < _start_render_index_x+_num_to_render_x; g++ )
    {
      //
      // first, get the min and max for this gene
      //
      tmp_min_max_v = cell_of_interest.getTimeSieresMinMax(g);
      min_max_v[0] = tmp_min_max_v[0];
      min_max_v[1] = tmp_min_max_v[1];
      for (int i = 0; i < group.getNumElements(); i++ )
      {
        ge = group.getElement(i);
        if ( !ge.render() ) continue;

        Embryo e = _embryos[embryoID2Index(ge.getEmbryoID())];
        tmp_min_max_v = e.getCell(e.cellID2Index(ge.getCellID())).getTimeSieresMinMax(g);
        min_max_v[0] = min( min_max_v[0], tmp_min_max_v[0] );
        min_max_v[1] = max( min_max_v[1], tmp_min_max_v[1] );
      }

      //
      // render the overlay for this gene
      //
      fill( CURVEMAP_VIEW_OVERLAY_PLOT_BACKGROUD_COLOR  );
      noStroke();
      rect( px, y, _overlay_plot_width, _overlay_plot_height );

      int ro_x=0, ro_y=0;
      boolean cg_cell_rolled = (!_render_comp_group_frame && _embryo_view.cellRolledOver() && 
        (_embryo_view.rolledOverCell() == _embryos[a.getSelectedEmbryo()].cellID2Index(cell_of_interest.getID())) && (cell_of_interest.getEmbryoID().equals(_embryos[a.getSelectedEmbryo()].getID())) );
      if ( !(_cell_rolled_over && (_rolled_over_cell == -1)) && !_coi_selected && !_comp_group_map.rolledOverCOI() && !cg_cell_rolled )
        renderLinePlot( cell_of_interest, px, y, g, min_max_v, 
        CURVEMAP_VIEW_OVERLAY_PLOT_LINE_COLOR, CURVEMAP_VIEW_OVERLAY_PLOT_LINE_WEIGHT );
      else
      {
        ro_x = px;
        ro_y = y;
      }
      int counter = 0, s_cell = 0;
      boolean loaded_group_rolled = false;
      boolean specific_loaded_group_rolled;
      for ( int i = 0; i < group.getNumElements(); i++ )
      {
        ge = group.getElement(i);
        if ( !ge.render() ) continue;

        cg_cell_rolled = (!_render_comp_group_frame && _embryo_view.cellRolledOverInCreatedGroup() && 
          (_embryo_view.rolledOverCell() == _embryos[a.getSelectedEmbryo()].cellID2Index(ge.getCellID())) && (ge.getEmbryoID().equals(_embryos[a.getSelectedEmbryo()].getID())) );
        specific_loaded_group_rolled = _save_load_view.loadedGroupRolledOver() && (_save_load_view.rolledOverLoadedGroup() == ge.existingGroupIndex());
        loaded_group_rolled = loaded_group_rolled || specific_loaded_group_rolled;
        if ( !(_cell_rolled_over && (_rolled_over_cell == counter)) && !(_comp_group_map.rolledOverGroupCell() && (_comp_group_map.groupCellRolledOver() == i)) && !cg_cell_rolled && !specific_loaded_group_rolled )
        {
          Embryo e = _embryos[embryoID2Index(ge.getEmbryoID())];
          renderLinePlot( e.getCell(e.cellID2Index(ge.getCellID())), px, y, g, min_max_v, 
          CURVEMAP_VIEW_OVERLAY_PLOT_LINE_COLOR, CURVEMAP_VIEW_OVERLAY_PLOT_LINE_WEIGHT );
        }
        else
        {
          ro_x = px;
          ro_y = y;
          s_cell = i;
        }

        ++counter;
      }

      // render the rolled over cell
      cg_cell_rolled = (!_render_comp_group_frame && _embryo_view.cellRolledOver() && 
        (_embryo_view.rolledOverCell() == _embryos[a.getSelectedEmbryo()].cellID2Index(cell_of_interest.getID())) && (cell_of_interest.getEmbryoID().equals(_embryos[a.getSelectedEmbryo()].getID())) );
      boolean cg_cell_rolled2 = (!_render_comp_group_frame && _embryo_view.cellRolledOverInCreatedGroup() && 
        (_embryo_view.rolledOverCell() == _embryos[a.getSelectedEmbryo()].cellID2Index(group.getElement(s_cell).getCellID())) && (group.getElement(s_cell).getEmbryoID().equals(_embryos[a.getSelectedEmbryo()].getID())) );

      if ( (_cell_rolled_over && (_rolled_over_cell == -1)) || _coi_selected || _comp_group_map.rolledOverCOI() || cg_cell_rolled )
        renderLinePlot( cell_of_interest, ro_x, ro_y, g, min_max_v, 
        CURVEMAP_VIEW_SELECTED_COLOR, CURVEMAP_VIEW_ROLLED_OVER_WEIGHT );
      if ( (_cell_rolled_over && (_rolled_over_cell != -1)) || cg_cell_rolled2 )
      {
        ge = group.getElement(s_cell);
        Embryo e = _embryos[embryoID2Index(ge.getEmbryoID())];
        renderLinePlot( e.getCell(e.cellID2Index(ge.getCellID())), ro_x, ro_y, g, min_max_v, 
        CURVEMAP_VIEW_OVERLAY_PLOT_ROLLED_OVERLINE_COLOR, CURVEMAP_VIEW_ROLLED_OVER_WEIGHT );
      }
      else if ( _comp_group_map.rolledOverGroupCell() )
      {
        ge = group.getElement(_comp_group_map.groupCellRolledOver());
        if ( ge.render() )
        {
          Embryo e = _embryos[embryoID2Index(ge.getEmbryoID())];
          renderLinePlot( e.getCell(e.cellID2Index(ge.getCellID())), ro_x, ro_y, g, min_max_v, 
          CURVEMAP_VIEW_OVERLAY_PLOT_ROLLED_OVERLINE_COLOR, CURVEMAP_VIEW_ROLLED_OVER_WEIGHT );
        }
      }
      else if ( loaded_group_rolled )
      {
        for ( int i = 0; i < group.getNumElements(); i++ )
        {
          ge = group.getElement(i);
          if ( !ge.render() ) continue;
          if ( !(_save_load_view.rolledOverLoadedGroup() == ge.existingGroupIndex()) ) continue;

          Embryo e = _embryos[embryoID2Index(ge.getEmbryoID())];
          renderLinePlot( e.getCell(e.cellID2Index(ge.getCellID())), px, y, g, min_max_v, 
          ((Group)_existing_groups.get(ge.existingGroupIndex())).getColor(), CURVEMAP_VIEW_ROLLED_OVER_WEIGHT );
        }
      }

      // outline the plots
      noFill();
      stroke( CURVEMAP_VIEW_OVERLAY_PLOT_OUTLINE_COLOR );
      strokeWeight( CURVEMAP_VIEW_OVERLAY_PLOT_OUTLINE_WEIGHT );
      rect( px, y, _filled_plot_width, _filled_plot_height );

      // write the min and max
//      textFont( _pixel_font_8 );
//      fill( CURVEMAP_VIEW_OVERLAY_PLOT_TEXT_COLOR );
//      textAlign( RIGHT, BOTTOM );
//      //text( min_max_v[1], tx_offset+px+_overlay_plot_width, ty_offset+y );
//      text( "1.0", tx_offset+px+_overlay_plot_width, ty_offset+y );
//      textAlign( RIGHT, TOP );
//      //text( min_max_v[0], tx_offset+px+_overlay_plot_width, ty_offset+y+_overlay_plot_height+2 );
//      text( "0.0", tx_offset+px+_overlay_plot_width, ty_offset+y+_overlay_plot_height+2 );

      px += _overlay_plot_offset + _overlay_plot_width;
    }

    // write the min and max
    px -= (_overlay_plot_offset + _overlay_plot_width - 3);
    textFont( _pixel_font_8 );
    fill( CURVEMAP_VIEW_OVERLAY_PLOT_TEXT_COLOR );
    textAlign( LEFT, TOP );
    //text( min_max_v[1], tx_offset+px+_overlay_plot_width, ty_offset+y );
    text( "1.0", tx_offset+px+_overlay_plot_width, ty_offset+y );
    textAlign( LEFT, BOTTOM );
    //text( min_max_v[0], tx_offset+px+_overlay_plot_width, ty_offset+y+_overlay_plot_height+2 );
    text( "0.0", tx_offset+px+_overlay_plot_width, ty_offset+y+_overlay_plot_height+2+y_bottom_offset );

  }

  void renderLinePlot( Cell c, int x, int y, int gene, float[] min_max_v, color stroke_color, int stroke_weight )
  {
    float v, next_v;
    float v_diff = min_max_v[1] - min_max_v[0];
    int vert_y, next_vert_y;
    int vert_x = x;
    for ( int i = 0; i < NUM_TIMESTEPS-1; i++ )
    {
      v = c.getProfileValue( gene, i );
      next_v = c.getProfileValue( gene, i+1 );

      // check if we have missing data for this interval
      if ( Float.isNaN(v) || Float.isNaN(next_v) ) continue;

      //vert_y = y + round( (1.0 - (v - min_max_v[0]) / v_diff) * (float)_filled_plot_height);
      //next_vert_y = y + round( (1.0 - (next_v - min_max_v[0]) / v_diff) * (float)_filled_plot_height);

      vert_y = y + round( (1.0 - min(v, 1.0)) * (float)_filled_plot_height);
      next_vert_y = y + round( (1.0 - min(next_v, 1.0)) * (float)_filled_plot_height);

      // render the interval line
      noFill();
      stroke( stroke_color );
      strokeWeight( stroke_weight );
      line( vert_x, vert_y, vert_x+_filled_plot_timestep_width, next_vert_y );

      vert_x += _filled_plot_timestep_width; 
    }
  }

  void renderCurvemapScrollBars()
  {
    int x, y;
    if ( _num_to_render_x != NUM_GENES )
    {
      noFill();
      stroke( SCROLL_BACKGROUND_COLOR );
      strokeWeight(  SCROLL_LINE_WEIGHT );
      line( _scroll_pos_x[0], _scroll_pos_x[1], _scroll_pos_x[0]+_scroll_length_x, _scroll_pos_x[1] );

      x = _scroll_pos_x[0];
      y = _scroll_pos_x[1]-_scroll_tick_width/2;
      line( x, y, x, y+_scroll_tick_width );
      x += _scroll_length_x;
      line( x, y, x, y+_scroll_tick_width );

      // and the scroll bar
      if ( !_over_x_scroll ) stroke( SCROLL_COLOR );
      else stroke( SCROLL_SELECTED_COLOR );
      strokeWeight(  SCROLL_BAR_LINE_WEIGHT );
      int min_pos = _scroll_pos_x[0] + _start_render_index_x*_scroll_tick_interval_x;
      int max_pos = _scroll_pos_x[0] + (_start_render_index_x+_num_to_render_x)*_scroll_tick_interval_x;
      line( min_pos, _scroll_pos_x[1], max_pos, _scroll_pos_x[1] );
    }

    if ( _num_to_render_y != _group_cells_rendered_counter )
    {
      noFill();
      stroke( SCROLL_BACKGROUND_COLOR );
      strokeWeight(  SCROLL_LINE_WEIGHT );
      line( _scroll_pos_y[0], _scroll_pos_y[1], _scroll_pos_y[0], _scroll_pos_y[1]+_scroll_length_y );

      x = _scroll_pos_y[0]-_scroll_tick_width/2;
      y = _scroll_pos_y[1];
      line( x, y, x+_scroll_tick_width, y );
      y += _scroll_length_y;
      line( x, y, x+_scroll_tick_width, y );

      // and the scroll bar
      if ( !_over_y_scroll ) stroke( SCROLL_COLOR );
      else stroke( SCROLL_SELECTED_COLOR );
      strokeWeight(  SCROLL_BAR_LINE_WEIGHT );
      int min_pos = _scroll_pos_y[1] + _start_render_index_y*_scroll_tick_interval_y;
      int max_pos = _scroll_pos_y[1] + (_start_render_index_y+_num_to_render_y)*_scroll_tick_interval_y;
      line( _scroll_pos_y[0], min_pos, _scroll_pos_y[0], max_pos );
    }
  }

  void mousePressedInView( int mx, int my, boolean right_pressed )
  {
    mx -= _frame_pos[0];
    my -= _frame_pos[1];

    if ( (my <= 0) && (my >= -_tab_height) )
    {
      if ( overTabs( mx, my ) )
      {
        setScrollYDimensions();
      }
    }

    else if ( (my >= _seperator_line_pos-4) && (my <= _seperator_line_pos+4) ) 
      _curvemap_divider_line_selected = true;

    else if ( (my >= _upper_pane_pos[1]+_comp_group_map_pos[1]) && (my <= _upper_pane_pos[1]+_comp_group_map_pos[1]+_upper_pane_height-5) )
    {
      if ( _render_comp_group_frame )
      {
        if ( _comp_group_map.mousePressedInView( mx-_upper_pane_pos[0]-_comp_group_map_pos[0], my-_upper_pane_pos[1]-_comp_group_map_pos[1], right_pressed ) )
        {
          setScrollYDimensions();
          int show_me = _comp_group_map.getPressedCellRenderNumber();
          if ( _num_to_render_y != _group_cells_rendered_counter )
          {
            if ( show_me < _start_render_index_y ) _start_render_index_y = show_me;
            else if ( show_me >= (_start_render_index_y+_num_to_render_y) ) _start_render_index_y = show_me-_num_to_render_y+1;
          }
        }
      }
      else
      {
        if ( _save_load_view.mousePressedInView( mx-_upper_pane_pos[0]-_save_load_view_pos[0], my-_upper_pane_pos[1]-_save_load_view_pos[1] ) )
          setScrollYDimensions();
      }
    }
    else if ( (mx >= (_lower_pane_pos[0]+_curvemap_pos[0] - 25)) && (mx <= (_lower_pane_pos[0]+_curvemap_pos[0])) )
    {
      boolean over_cell_id = overCellID( mx, my );
      if ( right_pressed && _render_comp_group_frame && over_cell_id )
      {
        Embryo e = _embryos[embryoID2Index(_rolled_over_cell_embryo_id)];
        if ( !e.getCell( _rolled_over_cell_index ).inCreatedGroup() )
        {
          e.getCell( _rolled_over_cell_index ).inCreatedGroup( true );
          GroupElement ge = new GroupElement( e.getID(), e.getCell( _rolled_over_cell_index ).getID() );
          ge.render( true );
          _created_group.addElement( ge );
        }
      }
      else if ( _rolled_over_cell == -1 )
      {
        _coi_selected = !_coi_selected;
      }
    }

    else if ( mx <= (_lower_pane_pos[0]+_curvemap_pos[0]-CURVEMAP_VIEW_PANE_BORDER-CURVEMAP_VIEW_INSET/2) )
    {
      if ( !overRemoveButton( mx-_lower_pane_pos[0]-_delete_column_pos[0], my-_lower_pane_pos[1]-_delete_column_pos[1] ) &&
        !overAllButton( mx-_lower_pane_pos[0]-_delete_column_pos[0], my-_lower_pane_pos[1]-_delete_column_pos[1] ) )
        overDeleteColumn( mx-_lower_pane_pos[0]-_delete_column_pos[0]-_delete_column_width/2, my-_lower_pane_pos[1]-_delete_column_pos[1] );
    }

    else if ( (my >= _lower_pane_pos[1]+_scroll_pos_x[1]-_scroll_tick_width/2) && (my <= _lower_pane_pos[1]+_scroll_pos_x[1]+_scroll_tick_width/2) )
    {
      if ( overXSlider( mx-_lower_pane_pos[0]-_scroll_pos_x[0], my-_lower_pane_pos[1]-_scroll_pos_x[1] ) )
      {
        _over_x_scroll = true;
        _over_x_scroll_pos = mx;
      }
    }
    else if ( (mx >= _lower_pane_pos[0]+_scroll_pos_y[0]-_scroll_tick_width/2) && (mx <= _lower_pane_pos[0]+_scroll_pos_y[0]+_scroll_tick_width/2) )
    {
      if ( overYSlider( mx-_lower_pane_pos[0]-_scroll_pos_y[0], my-_lower_pane_pos[1]-_scroll_pos_y[1] ) )
      {
        _over_y_scroll = true;
        _over_y_scroll_pos = my;
      }
    }
  }

  boolean overXSlider( int mx, int my )
  {
    int min_pos = _start_render_index_x*_scroll_tick_interval_x;
    int max_pos = (_start_render_index_x+_num_to_render_x)*_scroll_tick_interval_x;
    if ( (mx >= min_pos) && (mx <= max_pos) ) return true;

    return false;
  }

  boolean overYSlider( int mx, int my )
  {
    int min_pos = _start_render_index_y*_scroll_tick_interval_y;
    int max_pos = (_start_render_index_y+_num_to_render_y)*_scroll_tick_interval_y;
    if ( (my >= min_pos) && (my <= max_pos) ) return true;

    return false;
  }

  void mouseMovedInView( int mx, int my )
  {
    mx -= _frame_pos[0];
    my -= _frame_pos[1];

    _cell_rolled_over = false;
    _comp_group_map.reinitFlags();
    if ( (my >= _seperator_line_pos-4) && (my <= _seperator_line_pos+4) ) 
      _curvemap_divider_line_hoover = true;
    else if ( (my >= _upper_pane_pos[1]+_comp_group_map_pos[1]) && (my <= _upper_pane_pos[1]+_comp_group_map_pos[1]+_upper_pane_height) )
    {
      if ( _render_comp_group_frame )
        _comp_group_map.mouseMovedInView( mx-_upper_pane_pos[0]-_comp_group_map_pos[0], my-_upper_pane_pos[1]-_comp_group_map_pos[1] );
      else
        _save_load_view.mouseMovedInView( mx-_upper_pane_pos[0]-_save_load_view_pos[0], my-_upper_pane_pos[1]-_save_load_view_pos[1] );
    }
    else if ( (mx >= (_lower_pane_pos[0]+_curvemap_pos[0] - 25)) && (mx <= (_lower_pane_pos[0]+_curvemap_pos[0])) )
    {
      _cell_rolled_over = overCellID( mx, my );
    }
  }

  void mouseDraggedInView( int mx, int my, boolean right_pressed )
  {
    mx -= _frame_pos[0];
    my -= _frame_pos[1];

    if ( _over_x_scroll )
    {
      int movement = mx-_over_x_scroll_pos;

      if ( abs(movement) > _scroll_tick_interval_x/2 )
      {
        if ( movement < 0 )
        {
          if ( (_start_render_index_x-1) >= 0 )
          {
            _start_render_index_x = _start_render_index_x-1;
            _over_x_scroll_pos = mx;
          }
        } 
        else
        {
          if (  (_start_render_index_x+1+_num_to_render_x) <= NUM_GENES )
          {
            _start_render_index_x = _start_render_index_x+1;
            _over_x_scroll_pos = mx;
          }
        }
      }
    }

    else if ( _over_y_scroll )
    {
      int movement = my-_over_y_scroll_pos;

      if ( abs(movement) > _scroll_tick_interval_y/2 )
      {
        if ( movement < 0 )
        {
          if ( (_start_render_index_y-1) >= 0 )
          {
            _start_render_index_y = _start_render_index_y-1;
            _over_y_scroll_pos = my;
          }
        } 
        else
        {
          if (  (_start_render_index_y+1+_num_to_render_y) <= _group_cells_rendered_counter )
          {
            _start_render_index_y = _start_render_index_y+1;
            _over_y_scroll_pos = my;
          }
        }
      }
    }

    else if ( _curvemap_divider_line_selected )
    {
      _seperator_line_pos = min( max( my, CURVEMAP_VIEW_DIVIDER_LINE_MIN_MAX), _frame_height-2*CURVEMAP_VIEW_DIVIDER_LINE_MIN_MAX );
      setDimensions( _width, _height );
    }

    else if ( ((my >= _upper_pane_pos[1]+_comp_group_map_pos[1]) && (my <= _upper_pane_pos[1]+_comp_group_map_pos[1]+_upper_pane_height)) || (!_render_comp_group_frame && _save_load_view.loadSelected()) )
    {
      if ( _render_comp_group_frame )
        _comp_group_map.mouseDraggedInView( mx-_upper_pane_pos[0]-_comp_group_map_pos[0], my-_upper_pane_pos[1]-_comp_group_map_pos[1], right_pressed );
      else
        _save_load_view.mouseDraggedInView( mx-_upper_pane_pos[0]-_save_load_view_pos[0], my-_upper_pane_pos[1]-_save_load_view_pos[1] );
    }
  }

  void mouseReleasedSomewhere()
  {
    if ( _comp_group_map.mouseReleasedSomewhere() ) setScrollYDimensions();
    _over_x_scroll = _over_y_scroll = false;
    _save_load_view.mouseReleasedSomewhere();
  }

  void mousePressedSomewhere()
  {
    _save_load_view.mousePressedSomewhere();
  }

  boolean overAllButton( int mx, int my )
  {   
    // check if we are over the button
    int d_h = 3*_filled_plot_height/2 + CURVEMAP_VIEW_COI_BORDER + CURVEMAP_VIEW_SCROLL_OFFSET + SCROLL_WIDTH;
    d_h += (_filled_plot_height+CURVEMAP_VIEW_GROUP_CELL_BORDER)*_num_to_render_y;
    d_h -= (_filled_plot_height/3+CURVEMAP_VIEW_GROUP_CELL_BORDER);
    if ( (my >= d_h) && (my <= (d_h+CURVEMAP_VIEW_DELETE_BUTTON_HEIGHT)) )
    {
      Aggregation a = _aggregations[_selected_aggregation];
      Group group;
      if ( _render_comp_group_frame ) group = a.getComputationGroup(a.cellOfInterestIndex());
      else group = _created_group;
      GroupElement ge;
      for ( int i = 0; i < group.getNumElements(); i++ )
      {
        ge = group.getElement(i);
        if ( ge.render() ) ge.flagForRemoval( true );
      }
      _num_flagged_for_removal = _num_to_render_y;
      return true;
    }
    return false;
  }

  boolean overRemoveButton( int mx, int my )
  {   
    // check if we are over the button
    int d_h = 3*_filled_plot_height/2 + CURVEMAP_VIEW_COI_BORDER + CURVEMAP_VIEW_SCROLL_OFFSET + SCROLL_WIDTH;
    d_h += (_filled_plot_height+CURVEMAP_VIEW_GROUP_CELL_BORDER)*_num_to_render_y;
    d_h -= (_filled_plot_height/3+CURVEMAP_VIEW_GROUP_CELL_BORDER);
    d_h += 3*CURVEMAP_VIEW_DELETE_BUTTON_HEIGHT/2;
    if ( (my >= d_h) && (my <= (d_h+CURVEMAP_VIEW_DELETE_BUTTON_HEIGHT)) )
    {
      // remove the flagged cells
      Aggregation a = _aggregations[_selected_aggregation];
      Group group;
      if ( _render_comp_group_frame ) group = a.getComputationGroup(a.cellOfInterestIndex());
      else group = _created_group;
      GroupElement ge;
      for ( int i = group.getNumElements()-1; i >= 0; i-- )
      {
        ge = group.getElement(i);
        if ( ge.flaggedForRemoval() )
        {
          ge.flagForRemoval( false );
          ge.render( false );

          if ( !_render_comp_group_frame )
          { 
            Embryo e = _embryos[embryoID2Index(ge.getEmbryoID())];
            e.getCell(e.cellID2Index(ge.getCellID())).inCreatedGroup(false);
            group.removeElement(i);
          }
        }
      }
      _num_flagged_for_removal = 0;
      setScrollYDimensions();
      if ( !_render_comp_group_frame ) _save_load_view.setLoadedGroupsWidth();
      return true;
    }

    return false;
  }

  void overDeleteColumn( int mx, int my )
  { 
    if ( (mx <= -CURVEMAP_VIEW_DELETE_BUTTON_WIDTH/2) || (mx >= CURVEMAP_VIEW_DELETE_BUTTON_WIDTH/2) )
      return;

    GroupElement ge;
    Aggregation a = _aggregations[_selected_aggregation];
    Group group;
    if ( _render_comp_group_frame ) group = a.getComputationGroup(a.cellOfInterestIndex());
    else group = _created_group;
    int i = -1;
    int counter = -1;
    _num_flagged_for_removal = 0;
    int y = 3*_filled_plot_height/2 + CURVEMAP_VIEW_COI_BORDER + CURVEMAP_VIEW_SCROLL_OFFSET + SCROLL_WIDTH;
    while ( counter < _start_render_index_y )
    {
      ++i;
      ge = group.getElement(i);
      if ( ge.render() )
        ++counter;
    }

    while ( counter < _start_render_index_y+_num_to_render_y )
    {
      ge = group.getElement(i);
      ++i;
      if ( !ge.render() ) continue;

      if ( abs(my-y) <= CURVEMAP_VIEW_DELETE_BUTTON_WIDTH/2 )
      {
        if ( ge.flaggedForRemoval() )
        {
          ge.flagForRemoval( false );
          --_num_flagged_for_removal;
        }
        else
        {
          ge.flagForRemoval( true );
          ++_num_flagged_for_removal;
        }
        return;
      }
      y += _filled_plot_height + CURVEMAP_VIEW_GROUP_CELL_BORDER;
      ++counter;
    }
  }

  boolean overTabs( int mx, int my )
  {
    if ( (mx >= 0) && (mx <= _tab_width) )
    {
      _render_comp_group_frame = true;
      return true;
    }
    else if ( (mx >= (_tab_width+_tab_offset)) && (mx <= (2*_tab_width+_tab_offset)) )
    {
      _render_comp_group_frame = false;
      return true;
    }
    return false;
  }

  boolean overCellID( int mx, int my )
  { 
    int ny = _lower_pane_pos[1] + _curvemap_pos[1];

    // first check for the cell of interest
    if ( (my >= ny) && (my <= (ny+_filled_plot_height)) )
    {
      _rolled_over_cell = -1;

      Aggregation a = _aggregations[_selected_aggregation];
      GroupElement cell_of_interest_ge =  a.getCellOfInterest();
      _rolled_over_cell_index = _embryos[a.getSelectedEmbryo()].cellID2Index(cell_of_interest_ge.getCellID());
      _rolled_over_cell_embryo_id = cell_of_interest_ge.getEmbryoID();
      return true;
    }
    ny += _filled_plot_height + CURVEMAP_VIEW_COI_BORDER;

    // now check the cells in the group
    for ( int i = _start_render_index_y; i < _start_render_index_y+_num_to_render_y; i++ )
    {
      if ( (my >= ny) && (my <= (ny+_filled_plot_height)) )
      {
        _rolled_over_cell = i;
        if ( !_render_comp_group_frame )
        {
          _rolled_over_cell_embryo_id = _created_group.getElement(_rolled_over_cell).getEmbryoID();
          _rolled_over_cell_index = _embryos[embryoID2Index(_rolled_over_cell_embryo_id)].cellID2Index(_created_group.getElement(_rolled_over_cell).getCellID());
        }
        else
        {
          Aggregation a = _aggregations[_selected_aggregation];
          Group group = a.getComputationGroup(a.cellOfInterestIndex());
          GroupElement ge = group.getElement(0);
          int j = -1;
          int counter = -1;
          while ( counter < _rolled_over_cell )
          {
            ++j;
            ge = group.getElement(j);
            if ( ge.render() ) ++counter;
          }         

          _rolled_over_cell_embryo_id = ge.getEmbryoID();
          _rolled_over_cell_index = _embryos[embryoID2Index(_rolled_over_cell_embryo_id)].cellID2Index(ge.getCellID());
        }

        return true;
      }
      ny += _filled_plot_height + CURVEMAP_VIEW_GROUP_CELL_BORDER;
    }
    return false;
  }

  boolean keyPressedInView()
  {
    if ( !_render_comp_group_frame ) return _save_load_view.keyPressedInView();
    return false;
  }

}































