class EmbryoView
{
  int _width, _height;
  int _map_width, _map_height, _tick_length, _map_border; 
  int[] _map_pos;
  float _map_min_x, _map_max_x, _map_min_y, _map_max_y;

  int _slider_width, _slider_height, _slider_control_width, _slider_border;
  int[] _slider_pos, _slider_control_pos;
  float _slider_min, _slider_max, _slider_diff;
  float[] _abs_threshold, _rel_threshold;
  String _slider_min_string, _slider_max_string, _abs_threshold_min_string, _abs_threshold_max_string;
  PImage _slider_colorbar_img;
  int[] _slider_scent;
  int _min_scent_value, _max_scent_value;

  boolean _slider_min_control_pressed, _slider_max_control_pressed;
  boolean _cell_rolled_over;
  int _rolled_over_cell;

  int[] _x_scroll_pos, _y_scroll_pos;
  int _scroll_width, _scroll_offset, _x_scroll_length, _y_scroll_length;
  boolean _over_x_scroll, _over_y_scroll;
  int _over_x_scroll_pos, _over_y_scroll_pos;
  
  boolean _cell_rolled_over_in_created_group;

  EmbryoView()
  {
    _map_pos = new int[2];

    _tick_length = EMBRYO_TICK_LENGTH;
    _map_border = EMBRYO_MAP_BORDER;

    _slider_pos = new int[2];
    _slider_width = EMBRYO_VIEW_SLIDER_WIDTH;
    _slider_control_width = EMBRYO_VIEW_SLIDER_CONTROLS_WIDTH;
    _slider_border = EMBRYO_VIEW_SLIDER_BORDER;

    _map_min_x = _map_min_y = 0.0;
    _map_max_x = _map_max_y = 1.0;

    _abs_threshold = new float[2];
    _rel_threshold = new float[2];
    _slider_control_pos = new int[2];
    _rel_threshold[0] = 0.0;
    _rel_threshold[1] = 1.0;

    _cell_rolled_over = false;
    _rolled_over_cell = 0;

    _x_scroll_pos = new int[2];
    _y_scroll_pos = new int[2];
    _scroll_width = SCROLL_WIDTH;
    _scroll_offset = SCROLL_OFFSET;

    _over_x_scroll = _over_y_scroll = false;
    _over_x_scroll_pos = _over_y_scroll_pos = 0;
    
    _cell_rolled_over_in_created_group = false;
  }
  
  boolean cellRolledOver()
  {
    return _cell_rolled_over;
  }
  
  int rolledOverCell()
  {
    return _rolled_over_cell;
  }
  
  boolean cellRolledOverInCreatedGroup()
  {
    return _cell_rolled_over_in_created_group;
  }

  int getMaxWidth()
  {
    int v = EMBRYO_MAP_WIDTH + 2*_map_border + _slider_width + _slider_control_width + 2*_slider_border + _scroll_width + _scroll_offset;
    return v;
  }

  void setDimensions( int w, int h )
  {
    _y_scroll_pos[0] = 0;

    // the size and position of the embryo map 
    _map_pos[0] = _y_scroll_pos[0] + _scroll_width + _scroll_offset + 2*_map_border;
    _map_pos[1] = 2*_map_border;

    _y_scroll_pos[1] = _map_pos[1];
    _x_scroll_pos[0] = _map_pos[0];

    _map_width = min( w-2*_map_border-_slider_width-_slider_control_width-2*_slider_border-_scroll_width-_scroll_offset, EMBRYO_MAP_WIDTH );
    _map_height = min( h-3*_map_border-_scroll_width-_scroll_offset, EMBRYO_MAP_WIDTH );

    _x_scroll_pos[1] = _map_pos[1]+_map_height+_map_border+_scroll_offset;
    _y_scroll_length = _map_height;
    _x_scroll_length = _map_width;

    _slider_pos[0] = _map_pos[0]+_map_width+_slider_border;
    _slider_pos[1] = _map_pos[1];
    _slider_height = _map_height;

    setSliderControlPos();

    // set the min and max of the window
    _map_max_x = _map_min_x+(float)_map_width / (float)EMBRYO_MAP_WIDTH;
    if ( _map_max_x > 1.0 )
    {
      _map_min_x -= (_map_max_x-1.0);
      _map_max_x = 1.0;
    }
    _map_min_y = _map_max_y - ((float)_map_height / (float)EMBRYO_MAP_WIDTH);
    if ( _map_min_y < 0.0 )
    {
      _map_max_y += (0.0-_map_min_y);
      _map_min_y = 0.0;
    }

    // determine the size of the view
    _width = _map_width + 2*_map_border + _slider_width + _slider_control_width + 2*_slider_border + _scroll_width + _scroll_offset;
    _height = _map_height + 3*_map_border + _scroll_width + _scroll_offset;
  }

  void setSliderValues()
  {
    _slider_min = _aggregations[_selected_aggregation].getMinValue();
    _slider_max = _aggregations[_selected_aggregation].getMaxValue();
    _slider_diff = _slider_max - _slider_min;

    _slider_min_string = numToString( _slider_min );
    _slider_max_string = numToString( _slider_max );   

    setAbsThreshold();
    makeSliderScent();
    buildColorbar();
  }

  void setAbsThreshold()
  {
    _abs_threshold[0] = _slider_min + _slider_diff*_rel_threshold[0];
    _abs_threshold[1] = _slider_max - _slider_diff*(1.0 - _rel_threshold[1]);

    _abs_threshold_min_string = numToString( _abs_threshold[0] );
    _abs_threshold_max_string = numToString( _abs_threshold[1] );
  }

  void setSliderControlPos()
  {
    _slider_control_pos[0] = _slider_height - (int)(_rel_threshold[0]*(float)_slider_height);
    _slider_control_pos[1] = _slider_height - (int)(_rel_threshold[1]*(float)_slider_height);
  }

  void setRelThresholdMin()
  {
    _rel_threshold[0] = (float)(_slider_height - _slider_control_pos[0])/(float)_slider_height;
    _abs_threshold[0] = _slider_min + _slider_diff*_rel_threshold[0];
    _abs_threshold_min_string = numToString( _abs_threshold[0] );
  } 

  void setRelThresholdMax()
  {
    _rel_threshold[1] = (float)(_slider_height - _slider_control_pos[1])/(float)_slider_height;
    _abs_threshold[1] = _slider_max - _slider_diff*(1.0 - _rel_threshold[1]);
    _abs_threshold_max_string = numToString( _abs_threshold[1] );
  } 

  void buildColorbar()
  {
    _slider_colorbar_img = createImage( _slider_width, _slider_height, RGB );

    // set the colored pixels
    color c;
    int scent_diff = _max_scent_value - _min_scent_value;
    int num_to_scent;
    for ( int i = 0; i < _slider_height; i++ )
    {
      num_to_scent = ceil((float)(_slider_scent[(_slider_height-i-1)]-_min_scent_value)/(float)scent_diff*(float)_slider_width);
      c = EMBRYO_SLIDER_SCENT_COLOR;
      int s;
      for ( s = 0; s < num_to_scent; s++ )
        _slider_colorbar_img.pixels[i*_slider_width+s] = c;

      c = getColor( _aggregations[_selected_aggregation].getColorMap(), (1.0 - (float)i/(float)_slider_height) );
      for ( int j = s; j < _slider_width; j++ )
        _slider_colorbar_img.pixels[i*_slider_width+j] = c;
    }
  }

  void makeSliderScent()
  {
    // initialize the histogram
    _slider_scent = new int[_slider_height];
    for ( int i = 0; i < _slider_height; i++ )
      _slider_scent[i] = 0;

    // go through the cells and count
    _min_scent_value = 0;
    _max_scent_value = 0;
    float v_interval = 1.0 / (float)(_slider_height-1);
    Aggregation a = _aggregations[_selected_aggregation];
    for ( int i = 0; i < a.getNumCells(); i++ )
      ++_slider_scent[floor(a.getCell(i).getNormalizedValue()/v_interval)];

    // find the min and max
    for ( int i = 0; i < _slider_height; i++ )
      _max_scent_value = max( _max_scent_value, _slider_scent[i] );
  }


  int getViewWidth()
  { 
    return _width;
  }

  int getViewHeight()
  {
    return _height;
  }


  void render()
  {
    strokeCap(SQUARE);
    int x, y, offset, tx_offset=0, ty_offset=0;

    noSmooth();

    //
    // render the scroll bars
    //

    // y axis scroll bar
    if ( _map_height < EMBRYO_MAP_WIDTH )
    {
      strokeWeight( SCROLL_LINE_WEIGHT );
      stroke( SCROLL_BACKGROUND_COLOR );    
      line( _y_scroll_pos[0], _y_scroll_pos[1], _y_scroll_pos[0]+_scroll_width, _y_scroll_pos[1] );
      line( _y_scroll_pos[0]+_scroll_width/2, _y_scroll_pos[1], _y_scroll_pos[0]+_scroll_width/2, _y_scroll_pos[1]+_y_scroll_length );
      line( _y_scroll_pos[0], _y_scroll_pos[1]+_y_scroll_length, _y_scroll_pos[0]+_scroll_width, _y_scroll_pos[1]+_y_scroll_length );

      strokeWeight( SCROLL_BAR_LINE_WEIGHT );
      if ( !_over_y_scroll ) stroke( SCROLL_COLOR );
      else stroke( SCROLL_SELECTED_COLOR );
      int max_pos = _y_scroll_pos[1]+(int)((1.0-_map_max_y)*(float)_y_scroll_length);
      int min_pos = _y_scroll_pos[1]+(int)((1.0-_map_min_y)*(float)_y_scroll_length);
      line( _y_scroll_pos[0]+_scroll_width/2, max_pos, _y_scroll_pos[0]+_scroll_width/2, min_pos );
    }

    // x axis scroll bar
    if ( _map_width < EMBRYO_MAP_WIDTH )
    {
      strokeWeight( SCROLL_LINE_WEIGHT );
      stroke( SCROLL_BACKGROUND_COLOR ); 
      line( _x_scroll_pos[0], _x_scroll_pos[1], _x_scroll_pos[0], _x_scroll_pos[1]+_scroll_width );
      line( _x_scroll_pos[0], _x_scroll_pos[1]+_scroll_width/2, _x_scroll_pos[0]+_x_scroll_length, _x_scroll_pos[1]+_scroll_width/2 );
      line( _x_scroll_pos[0]+_x_scroll_length, _x_scroll_pos[1], _x_scroll_pos[0]+_x_scroll_length, _x_scroll_pos[1]+_scroll_width );

      strokeWeight( SCROLL_BAR_LINE_WEIGHT );
      if ( !_over_x_scroll ) stroke( SCROLL_COLOR );
      else stroke( SCROLL_SELECTED_COLOR );
      int min_pos = _x_scroll_pos[0]+(int)(_map_min_x*(float)_x_scroll_length);
      int max_pos = _x_scroll_pos[0]+(int)(_map_max_x*(float)_x_scroll_length);
      line( min_pos, _x_scroll_pos[1]+_scroll_width/2, max_pos, _x_scroll_pos[1]+_scroll_width/2 );
    }
    
    //
    // render the title
    //
    if( _use_text_offset )
    {
      tx_offset = _embryo_view_origin[0];
      ty_offset = _embryo_view_origin[1];
    }
    int y_bottom_offset = 0;
    if ( !_use_text_offset )
    { 
      ty_offset = -PDF_OFFSET;
      y_bottom_offset = PDF_OFFSET-1;
    }
    
    x = _map_pos[0]-45;
    y = 2;
    textFont( _verdana_font_16 );
    fill( TITLE_COLOR );
    textAlign( LEFT, BOTTOM );
    text( "Embryo Map", x+tx_offset, y+ty_offset+y_bottom_offset );
//    strokeWeight( 1 );
//    stroke(  KEY_COLOR );
//    //y -= 24;
//    x = _map_pos[0]-30;
//    line( x, y, textWidth("Unrolled View")-45, y );
//    line( x, y, x, _height );


    //
    // render the embryo map
    //
    if( _use_text_offset )
    {
      tx_offset = _embryo_view_origin[0] + _map_pos[0];
      ty_offset = _embryo_view_origin[1] + _map_pos[1];
    }
    

    pushMatrix();    
    translate( _map_pos[0], _map_pos[1] );

    fill( EMBRYO_MAP_BACKGROUND_COLOR );
    stroke( EMBRYO_MAP_OUTLINE_COLOR );
    strokeWeight( EMBRYO_MAP_OUTLINE_WEIGHT );
    rectMode( CORNER );
    rect( 0, 0, _map_width, _map_height );

    //
    // make the tick marks (assumes a domain of [0,1]).
    //  also assumes just 3 ticks per axis for now....
    //
    textFont(_pixel_font_8); 
    fill( EMBRYO_VIEW_TEXT_COLOR );
    stroke( EMBRYO_MAP_OUTLINE_COLOR );
    strokeWeight( EMBRYO_MAP_OUTLINE_WEIGHT );

    // the x-axis ticks
    textAlign( CENTER, TOP );
    offset = 1;
    x = 0;
    y = _map_height;
    line( x, y, x, y+_tick_length );
    String s = nf( _map_min_x, 1, 1 );
    text( s, x+tx_offset, y+ty_offset+_tick_length+offset+2 );

    x += _map_width/2;
    if ( (_map_max_x-_map_min_x) > 0.5 )
    {
      s = nf( (_map_max_x-_map_min_x)*0.5+_map_min_x, 1, 1 );
      line( x, y, x, y+_tick_length );
      text( s, x+tx_offset, y+ty_offset+_tick_length+offset+2 );
    }

    x += _map_width/2;
    line( x, y, x, y+_tick_length );
    s = nf( _map_max_x, 1, 1 );
    text( s, x+tx_offset, y+ty_offset+_tick_length+offset+2 );

    // the y-axis ticks    
    textAlign( RIGHT, CENTER );
    x = 0;
    y = _map_height;
    line( x, y, x-_tick_length, y );
    s = nf( _map_min_y, 1, 1 );
    text( s, x+tx_offset-_tick_length-offset-2, y+ty_offset );

    y -= _map_height/2;
    if ( (_map_max_y-_map_min_y) > 0.5 )
    {
      line( x, y, x-_tick_length, y );
      s = nf( (_map_max_y-_map_min_y)*0.5+_map_min_y, 1, 1 );
      text( s, x+tx_offset-_tick_length-offset-2, y+ty_offset );
    }

    y -= _map_height/2;
    line( x, y, x-_tick_length, y );
    s = nf( _map_max_y, 1, 1 );
    text( s, x+tx_offset-_tick_length-offset-2, y+ty_offset );

    //
    // render the cells
    //
    noFill();
    strokeWeight( EMBRYO_VIEW_CELL_STROKE_WEIGHT );
    float[] pos;
    float v;
    boolean in_threshold;
    Aggregation a = _aggregations[_selected_aggregation];
    Embryo e = _embryos[a.getSelectedEmbryo()];
    GroupElement ge;
    Cell c;
    int cell_of_interest_index = a.cellOfInterestIndex();
    int cell_width;
    boolean cell_name_rolled_over = (_curvemap_view.cellRolledOver() && !_curvemap_view.renderCompGroupFrame());
    String cm_rolled_embryo_id = _curvemap_view.rolledOverCellEmbryoID(); 
    int cell_name_rolled_over_index = _curvemap_view.rolledOverCellIndex();
    for ( int i = 0; i < a.getNumCells(); i++ )
    {      
      ge = a.getCell( i );      
      c = e.getCell(e.cellID2Index(ge.getCellID()));

      // check for rolled over -- gets rendered last!
      if ( (_cell_rolled_over && (i == _rolled_over_cell)) || c.inCreatedGroup() /*|| 
        (_curvemap_view.selectedCellRolledOver() && (i==cell_of_interest_index)) || 
        (cell_name_rolled_over && (cell_name_rolled_over_index == i) && (cm_rolled_embryo_id.equals(e.getID())))*/  )
        continue;

      v = ge.getValue();
      pos = c.getPos();

      if ( (v < _abs_threshold[0] ) || (v > _abs_threshold[1]) ) in_threshold = false;
      else in_threshold = true;

      if ( (pos[0] >= _map_min_x) && (pos[0] <= _map_max_x) && (pos[1] >= _map_min_y) && (pos[1] <= _map_max_y) )
      {
        x = (int)((pos[0]-_map_min_x)*(float)EMBRYO_MAP_WIDTH);
        y = (int)((_map_max_y-pos[1])*(float)EMBRYO_MAP_WIDTH); // translate to bottom of map....

        cell_width = EMBRYO_VIEW_CELL_WIDTH;

        if ( i == cell_of_interest_index )
        {
          fill( EMBRYO_VIEW_COI_COLOR );
          stroke( EMBRYO_VIEW_COI_COLOR );
        }
        else if ( in_threshold )
        {
          fill( getColor(a.getColorMap(), ge.getNormalizedValue()) );
          stroke( getColor(a.getColorMap(), ge.getNormalizedValue()) );
        }
        else
        {
          noFill();
          stroke( getColor(a.getColorMap(), ge.getNormalizedValue()) );
        }

        ellipse( x, y, cell_width, cell_width );
      }
    }

    rectMode( CENTER );
    for ( int i = 0; i < a.getNumCells(); i++ )
    {      
      ge = a.getCell( i );      
      c = e.getCell(e.cellID2Index(ge.getCellID()));

      // check for rolled over -- gets rendered last!
      if ( (_cell_rolled_over && (i == _rolled_over_cell)) || !c.inCreatedGroup() || 
        (cell_name_rolled_over && (cell_name_rolled_over_index == i) && (cm_rolled_embryo_id.equals(e.getID()))) )
        continue;

      v = ge.getValue();
      pos = c.getPos();

      if ( (v < _abs_threshold[0] ) || (v > _abs_threshold[1]) ) in_threshold = false;
      else in_threshold = true;

      if ( (pos[0] >= _map_min_x) && (pos[0] <= _map_max_x) && (pos[1] >= _map_min_y) && (pos[1] <= _map_max_y) )
      {
        x = (int)((pos[0]-_map_min_x)*(float)EMBRYO_MAP_WIDTH);
        y = (int)((_map_max_y-pos[1])*(float)EMBRYO_MAP_WIDTH); // translate to bottom of map....

        cell_width = EMBRYO_VIEW_CELL_WIDTH;
        strokeWeight( EMBRYO_VIEW_CELL_STROKE_WEIGHT );

        if ( i == cell_of_interest_index )
        {
          fill( EMBRYO_VIEW_COI_COLOR );
          stroke( EMBRYO_VIEW_IN_CREATED_GROUP_LINE_COLOR );
        }
        else if ( in_threshold )
        {
          fill( getColor(a.getColorMap(), ge.getNormalizedValue()) );
          stroke( EMBRYO_VIEW_IN_CREATED_GROUP_LINE_COLOR );
        }
        else
        {
          fill( EMBRYO_VIEW_IN_CREATED_GROUP_FILL_COLOR );
          stroke( getColor(a.getColorMap(), ge.getNormalizedValue()) );
        }

        //        if ( i == cell_of_interest_index ) fill( EMBRYO_VIEW_COI_COLOR );
        //        else fill( EMBRYO_VIEW_IN_CREATED_GROUP_FILL_COLOR );
        //        stroke( EMBRYO_VIEW_IN_CREATED_GROUP_LINE_COLOR );

        rect( x, y, cell_width+2, cell_width+2 );
      }
    }
    rectMode( CORNER );

    // if a cell is rolled over, render it on top
    if ( _cell_rolled_over )
    {
      ge = a.getCell( _rolled_over_cell );
      c = e.getCell(e.cellID2Index(ge.getCellID()));
      v = ge.getValue();
      pos = e.getCell(e.cellID2Index(ge.getCellID())).getPos();

      if ( (v < _abs_threshold[0] ) || (v > _abs_threshold[1]) ) in_threshold = false;
      else in_threshold = true;

      x = (int)((pos[0]-_map_min_x)*(float)EMBRYO_MAP_WIDTH);
      y = (int)((_map_max_y-pos[1])*(float)EMBRYO_MAP_WIDTH); // translate to bottom of map....

      cell_width = EMBRYO_VIEW_ROLLED_OVER_CELL_WIDTH;
      strokeWeight( EMBRYO_VIEW_ROLLED_OVER_CELL_WEIGHT );

      if ( !c.inCreatedGroup() )
      {
        stroke( EMBRYO_VIEW_ROLLOVER_COLOR );

        if ( _rolled_over_cell == cell_of_interest_index )
          fill( EMBRYO_VIEW_COI_COLOR );
        else if ( in_threshold )
          fill( getColor(a.getColorMap(), ge.getNormalizedValue()) );
        else
          fill( EMBRYO_VIEW_IN_CREATED_GROUP_FILL_COLOR );

        ellipse( x, y, cell_width, cell_width );
      }
      else
      {       
        stroke( EMBRYO_VIEW_IN_CREATED_GROUP_LINE_COLOR );

        //        if ( _rolled_over_cell == cell_of_interest_index ) fill( EMBRYO_VIEW_COI_COLOR );
        //        else fill( EMBRYO_VIEW_IN_CREATED_GROUP_FILL_COLOR );

        if ( _rolled_over_cell == cell_of_interest_index )
          fill( EMBRYO_VIEW_COI_COLOR );
        else if ( in_threshold )
          fill( getColor(a.getColorMap(), ge.getNormalizedValue()) );
        else
          fill( EMBRYO_VIEW_IN_CREATED_GROUP_FILL_COLOR );

        rectMode( CENTER );
        rect( x, y, cell_width+2, cell_width+2 );
        rectMode( CORNER );
      }
    }
    else if ( _curvemap_view.selectedCellRolledOver() )
    {
      ge = a.getCell( cell_of_interest_index );
      c = e.getCell(e.cellID2Index(ge.getCellID()));
      v = ge.getValue();
      pos = e.getCell(e.cellID2Index(ge.getCellID())).getPos();

      x = (int)((pos[0]-_map_min_x)*(float)EMBRYO_MAP_WIDTH);
      y = (int)((_map_max_y-pos[1])*(float)EMBRYO_MAP_WIDTH); // translate to bottom of map....

      cell_width = EMBRYO_VIEW_ROLLED_OVER_CELL_WIDTH;
      strokeWeight( EMBRYO_VIEW_ROLLED_OVER_CELL_WEIGHT );

      if ( !c.inCreatedGroup() )
      {
        stroke( EMBRYO_VIEW_ROLLOVER_COLOR );
        fill( EMBRYO_VIEW_COI_COLOR );

        ellipse( x, y, cell_width, cell_width );
      }
      else
      {       
        stroke( EMBRYO_VIEW_IN_CREATED_GROUP_LINE_COLOR );
        fill( EMBRYO_VIEW_COI_COLOR );

        rectMode( CENTER );
        rect( x, y, cell_width+2, cell_width+2 );
        rectMode( CORNER );
      }
    }
    else if ( cell_name_rolled_over && (cm_rolled_embryo_id.equals(e.getID())) )
    {
      ge = a.getCell( cell_name_rolled_over_index );
      c = e.getCell(e.cellID2Index(ge.getCellID()));
      v = ge.getValue();
      pos = e.getCell(e.cellID2Index(ge.getCellID())).getPos();

      x = (int)((pos[0]-_map_min_x)*(float)EMBRYO_MAP_WIDTH);
      y = (int)((_map_max_y-pos[1])*(float)EMBRYO_MAP_WIDTH); // translate to bottom of map....

      if ( (v < _abs_threshold[0] ) || (v > _abs_threshold[1]) ) in_threshold = false;
      else in_threshold = true;

      cell_width = EMBRYO_VIEW_ROLLED_OVER_CELL_WIDTH;
      strokeWeight( EMBRYO_VIEW_ROLLED_OVER_CELL_WEIGHT );

      if ( in_threshold )
      {
        fill( getColor(a.getColorMap(), ge.getNormalizedValue()) );
        stroke( EMBRYO_VIEW_IN_CREATED_GROUP_LINE_COLOR );
      }
      else
      {
        fill( EMBRYO_VIEW_IN_CREATED_GROUP_FILL_COLOR );
        stroke( getColor(a.getColorMap(), ge.getNormalizedValue()) );
      }

      rectMode( CENTER );
      rect( x, y, cell_width+2, cell_width+2 );
      rectMode( CORNER );
    }

    //
    // render the map labels
    //
    textFont(_pixel_font_8); 

    x = 0;
    y = -2+y_bottom_offset;
    fill( EMBRYO_VIEW_COI_COLOR );
    textAlign( LEFT, BOTTOM );
    text( "SELECTED CELL: " + a.getCellOfInterest().getCellID(), x+tx_offset, y+ty_offset );

    y -= 12;   
    fill( EMBRYO_VIEW_TEXT_COLOR );    
    text( "EMBRYO: " + e.getID(), x+tx_offset, y+ty_offset ); 

    x = _map_width;
    y = -2+y_bottom_offset;    
    textAlign( RIGHT, BOTTOM );
    text( "AGGREGATION: " + a.getAggID(), x+tx_offset, y+ty_offset );
    
    y -= 12;
    //x += _slider_border;
    //textAlign( LEFT, BOTTOM );
    text( "METRIC: " + a.getID(), x+tx_offset, y+ty_offset );
    
//    y -= 6;
//    x += _slider_border+_slider_width+35;
//    textAlign( RIGHT, BOTTOM );
//    text( "computation", x+tx_offset, y+ty_offset );
//    
//    y += 12;
//    text( "value", x+tx_offset, y+ty_offset );

    //
    // finished with the embryo map
    //
    popMatrix();

    //
    // render the slider
    //
    if( _use_text_offset )
    {
      tx_offset = _embryo_view_origin[0] + _slider_pos[0];
      ty_offset = _embryo_view_origin[1] + _slider_pos[1];
    }
    //else ty_offset -= PDF_OFFSET;

    pushMatrix();    
    translate( _slider_pos[0], _slider_pos[1] );

    textFont(_pixel_font_8); 

    // render the colormap
    image( _slider_colorbar_img, 0, 0, _slider_width, _slider_height );

    // outline the slider bar
    stroke( EMBRYO_MAP_OUTLINE_COLOR );
    strokeWeight( EMBRYO_MAP_OUTLINE_WEIGHT );
    noFill();
    rect( 0, 0, _slider_width, _slider_height );

    // render the controls    
    x = _slider_width;
    y = _slider_control_pos[0];
    if ( !_slider_min_control_pressed ) 
    {
      fill( EMBRYO_MAP_SLIDER_CONTROL_COLOR );
      stroke( EMBRYO_MAP_SLIDER_CONTROL_COLOR );
    } 
    else
    {
      fill( EMBRYO_MAP_SLIDER_CONTROL_SELECTED_COLOR );
      stroke( EMBRYO_MAP_SLIDER_CONTROL_SELECTED_COLOR );
    }
    triangle( x, y, x+_slider_control_width, y, x+_slider_control_width, y-_slider_control_width );
    textAlign( LEFT, BOTTOM );
    text( _abs_threshold_min_string, x+tx_offset+_slider_control_width+2, y+ty_offset+1+y_bottom_offset );

    y = _slider_control_pos[1];
    if ( !_slider_max_control_pressed ) 
    {
      fill( EMBRYO_MAP_SLIDER_CONTROL_COLOR );
      stroke( EMBRYO_MAP_SLIDER_CONTROL_COLOR );
    } 
    else
    {
      fill( EMBRYO_MAP_SLIDER_CONTROL_SELECTED_COLOR );
      stroke( EMBRYO_MAP_SLIDER_CONTROL_SELECTED_COLOR );
    }
    triangle( x, y, x+_slider_control_width, y, x+_slider_control_width, y+_slider_control_width );
    textAlign( LEFT, TOP );
    text( _abs_threshold_max_string, x+tx_offset+_slider_control_width+2, y+ty_offset+2 );

    // write the min and max values
    fill( EMBRYO_VIEW_TEXT_COLOR );
    stroke( EMBRYO_MAP_OUTLINE_COLOR );
    strokeWeight( EMBRYO_MAP_OUTLINE_WEIGHT );

    x = _slider_width;
    y = 0;
    line( x, y, x+_tick_length, y );
    textAlign( LEFT, BOTTOM );
    text( _slider_max_string, x+tx_offset+_tick_length, y+ty_offset+y_bottom_offset );

    y += _slider_height;
    line( x, y, x+_tick_length, y );
    textAlign( LEFT, TOP );
    text( _slider_min_string, x+tx_offset+_tick_length, y+ty_offset+2 );

    //
    // finished with the slider
    //
    popMatrix();

  }

  void mousePressedInView( int mx, int my, boolean ctrl_pressed )
  {
    if ( (mx >= _y_scroll_pos[0]) && (mx <= (_y_scroll_pos[0]+_scroll_width)) )
    {
      if ( overYSlider( mx-_y_scroll_pos[0], my-_y_scroll_pos[1] ) )
      {
        _over_y_scroll = true;
        _over_y_scroll_pos = my;
      }
    }
    else if ( (my >= _x_scroll_pos[1]) && (my <= (_x_scroll_pos[1]+_slider_width)) )
    {
      if ( overXSlider( mx-_x_scroll_pos[0], my-_x_scroll_pos[1] ) )
      {
        _over_x_scroll = true;
        _over_x_scroll_pos = mx;
      }
    }
    else if ( (mx >= _slider_pos[0]+_slider_width) && (mx <= _slider_pos[0]+_slider_width+_slider_control_width) )
    { 
      _slider_min_control_pressed = overSliderMinControl( mx, my );
      _slider_max_control_pressed = overSliderMaxControl( mx, my );
    } 

    else
    {
      int rv = overCell( mx-_map_pos[0], my-_map_pos[1] );
      if ( rv != -1 )
      {
        if ( !ctrl_pressed )
          _aggregations[_selected_aggregation].cellOfInterestIndex( rv );
        else
        {
          Aggregation a = _aggregations[_selected_aggregation];
          Embryo e = _embryos[a.getSelectedEmbryo()];
          //Cell c = e.getCell( rv );
          if ( !e.getCell( rv ).inCreatedGroup() )
          {
            e.getCell( rv ).inCreatedGroup( true );
            GroupElement ge = new GroupElement( e.getID(), e.getCell( rv ).getID() );
            ge.render( true );
            _created_group.addElement( ge );
          }
        }
      }
    }  
  }

  void cellRolledOverReinit()
  {
    _cell_rolled_over = false;
    _cell_rolled_over_in_created_group = false;
  }

  boolean mouseMovedInView( int mx, int my )
  {        
    cellRolledOverReinit();
    if ( (mx >= _map_pos[0]) && (mx <= _map_pos[0]+_map_width) && (my >= _map_pos[1]) && (my <= _map_pos[1]+_map_height) )
    {
      _rolled_over_cell = overCell( mx-_map_pos[0], my-_map_pos[1] );
      if ( _rolled_over_cell != -1 )
      {
        _cell_rolled_over = true;   
        
        Aggregation a = _aggregations[_selected_aggregation];
        if ( _embryos[a.getSelectedEmbryo()].getCell(_rolled_over_cell).inCreatedGroup() ) _cell_rolled_over_in_created_group = true;
      }

      return true;  
    }
    return false;
  }



  void mouseDraggedInView( int mx, int my, boolean right_pressed )
  {
    if ( _slider_min_control_pressed )
    {
      my -= _slider_pos[1];
      _slider_control_pos[0] = max( _slider_control_pos[1], min( _slider_height, my ) );
      setRelThresholdMin();
    }
    else if ( _slider_max_control_pressed )
    {
      my -= _slider_pos[1];
      _slider_control_pos[1] = min( _slider_control_pos[0], max( 0, my ) );
      setRelThresholdMax();
    }
    else if ( _over_y_scroll )
    {
      float movement = (float)(my - _over_y_scroll_pos)/(float)_y_scroll_length;
      _map_max_y -= movement;
      _map_min_y -= movement;

      if ( _map_max_y > 1.0 )
      {
        movement = _map_max_y - 1.0;
        _map_max_y = 1.0;
        _map_min_y -= movement;
      }
      if ( _map_min_y < 0.0 )
      {
        movement = 0.0-_map_min_y;
        _map_min_y = 0.0;
        _map_max_y += movement;
      }

      _over_y_scroll_pos = my;
    }
    else if ( _over_x_scroll )
    {
      float movement = (float)(mx-_over_x_scroll_pos)/(float)_x_scroll_length;
      _map_max_x += movement;
      _map_min_x += movement;

      if ( _map_max_x > 1.0 )
      {
        movement = _map_max_x - 1.0;
        _map_max_x = 1.0;
        _map_min_x -= movement;
      }
      if ( _map_min_x < 0.0 )
      {
        movement = 0.0-_map_min_x;
        _map_min_x = 0.0;
        _map_max_x += movement;
      }

      _over_x_scroll_pos = mx;
    }
    else if ( right_pressed )
    {
      cellRolledOverReinit();
      if ( (mx >= _map_pos[0]) && (mx <= _map_pos[0]+_map_width) && (my >= _map_pos[1]) && (my <= _map_pos[1]+_map_height) )
      {
        _rolled_over_cell = overCell( mx-_map_pos[0], my-_map_pos[1] );
        if ( _rolled_over_cell != -1 )
        {
          _cell_rolled_over = true; 

          Aggregation a = _aggregations[_selected_aggregation];
          Embryo e = _embryos[a.getSelectedEmbryo()];
          Cell c = e.getCell( _rolled_over_cell );
          if ( !c.inCreatedGroup() )
          {
            c.inCreatedGroup( true );
            GroupElement ge = new GroupElement( e.getID(), c.getID() );
            ge.render( true );
            _created_group.addElement( ge );
          }
        }
      }
    }
  }

  void mouseReleasedSomewhere()
  {
    _slider_min_control_pressed = _slider_max_control_pressed = false;
    _over_y_scroll = _over_x_scroll = false;
  }

  boolean overYSlider( int mx, int my )
  {
    int min_pos = (int)((1.0-_map_min_y)*(float)_y_scroll_length);
    int max_pos = (int)((1.0-_map_max_y)*(float)_y_scroll_length);
    if ( (my >= max_pos) && (my <= min_pos) ) return true;

    return false;
  }

  boolean overXSlider( int mx, int my )
  {
    int min_pos = (int)(_map_min_x*(float)_x_scroll_length);
    int max_pos = (int)(_map_max_x*(float)_x_scroll_length);
    if ( (mx >= min_pos) && (mx <= max_pos) ) return true;

    return false;
  }

  boolean overSliderMinControl( int mx, int my )
  {
    my -= _slider_pos[1];
    if ( (my <= _slider_control_pos[0]) && (my >= _slider_control_pos[0]-_slider_control_width) )
      return true;
    return false;
  }

  boolean overSliderMaxControl( int mx, int my )
  {
    my -= _slider_pos[1];
    if ( (my >= _slider_control_pos[1]) && (my <= _slider_control_pos[1]+_slider_control_width) )
      return true;
    return false;
  }

  int overCell( int mx, int my )
  {
    Aggregation a = _aggregations[_selected_aggregation];
    Embryo e = _embryos[a.getSelectedEmbryo()];
    float pos[] = new float[2];
    for ( int i = 0; i < a.getNumCells(); i++ )
    {
      pos = e.getCell(e.cellID2Index(a.getCell(i).getCellID())).getPos();
      if ( (abs((pos[0]-_map_min_x)*(float)EMBRYO_MAP_WIDTH - mx) <= (EMBRYO_VIEW_CELL_HALF_WIDTH+1)) && 
        (abs((_map_max_y-pos[1])*(float)EMBRYO_MAP_WIDTH - my) <= (EMBRYO_VIEW_CELL_HALF_WIDTH+1)) )
        return i;
    }

    return -1;
  }
}

































