//
// CHECK FOR EXISTANCE OF AGGREGATION FOR EACH EMBRYO
//
void checkExistanceOfAggregations()
{
  String e_id;
  int c_id, e_num_cells;
  for ( int i = 0; i < NUM_EMBRYOS; i++ )
  {
    // check if an aggregation exists for this embryo
    boolean exists = false;
    e_id = _embryos[i].getID();
    for ( int j = 0; j < NUM_AGGREGATIONS; j++ )
      if ( _aggregations[j].getEmbryoID().equals(e_id) ) exists = true;

    // if it doesn't exist, make a default
    if ( !exists )
    { 
      println( "Creating a default aggregation for embryo " + e_id );

      e_num_cells = _embryos[i].getNumCells();
      Aggregation a = new Aggregation( "default", e_id, "none", e_num_cells );
      GroupElement ge;
      Group cg;
      for ( int j = 0; j < e_num_cells; j++ )
      {
        c_id = _embryos[i].getCell(j).getID();
        ge = new GroupElement( e_id, c_id, Float.NaN );
        cg = new Group( "no computation group for cell " + c_id + " in embryo " + e_id );

        a.addCell( ge, cg );
      }

      // copy the agg array to a temporary array      
      Aggregation[] temp = new Aggregation[NUM_AGGREGATIONS];
      for ( int j = 0; j < NUM_AGGREGATIONS; j++ ) temp[j] = new Aggregation( _aggregations[j] );

      // now remake the agg array
      ++NUM_AGGREGATIONS;
      _aggregations = new Aggregation[NUM_AGGREGATIONS];
      for ( int j = 0; j < NUM_AGGREGATIONS-1; j++ ) _aggregations[j] = new Aggregation( temp[j] );
      _aggregations[NUM_AGGREGATIONS-1] = new Aggregation( a );
    }
  }
}

//
// SOME INITIALIZATIONS
//
void initializations()
{
  for ( int i = 0; i < NUM_AGGREGATIONS; i++ )
  {
    _aggregations[i].normalizeValues();
    _aggregations[i].initializeToRenderInGroups();
  }

  _embryo_view.setSliderValues();
  _aggregation_view.createThumbnails();
  _aggregation_view.setScrollDimensions();
  _curvemap_view.initializations();
  _curvemap_view.setScrollDimensions();

  // initialize the existing groups
  Group g;
  GroupElement ge;
  String[] rows;
  String[] cols;
  for ( int i = 0; i < _initial_group_files.size(); i++ )
  {
    rows = loadStrings( (String)_initial_group_files.get(i) );
    int counter = 0;
    while ( splitTokens(rows[counter],",")[0].substring(0,2).equals( "##" ) )
      ++counter;

    String id = splitTokens(rows[counter],",:")[1].trim();
    int num_ge = rows.length - counter - 2;

    println( "Created the group " + id + "." );

    counter += 2;
    g = new Group( id );
    for ( int j = 0; j < num_ge; j++ )
    {
      cols = splitTokens(rows[counter+j],",");
      String e_id = cols[0].trim();
      int c_id = parseInt( cols[1] );

      ge = new GroupElement( e_id, c_id );
      ge.fromAnExistingGroup( true );
      ge.existingGroupIndex( i );
      g.addElement( ge );
    }
    g.setColor( COLORMAP_CATAGORICAL[i % (COLORMAP_CATAGORICAL.length)] );
    _existing_groups.add( new Group(g) );
  }
}

//
// COLORMAPPING
//
color getColor( color[] cmap, float value )
{
  int num_intervals = cmap.length-1;
  float interval_size = 1.0/(float)num_intervals;
  float top_v;
  if ( !_invert_colormap )
  {
    for ( int i = 0; i < num_intervals; i++ )
    {
      float v = (float)(i+1)*interval_size;
      if ( value < v ) 
        return lerpColor( cmap[i], cmap[i+1], (value-(float)i*interval_size)/interval_size );
    }
  }
  else
  {
    for ( int i = 0; i < num_intervals; i++ )
    {
      float v = (float)(i+1)*interval_size;
      if ( value < v ) 
        return lerpColor( cmap[(num_intervals-i)-1], cmap[(num_intervals-i)], 1.0-(value-(float)i*interval_size)/interval_size );
    }
  }

  return cmap[cmap.length-1];
}

//
// CONVERTING NUMBERS TO STRINGS
//
String numToString( float v )
{
  int digis=5;
  String s;
  if ( abs(v) < 10 ) digis = 1;
  else if ( abs(v) < 100 ) digis = 2;
  else if ( abs(v) < 1000 ) digis = 3;
  else if ( abs(v) < 10000 ) digis = 4;
  if ( digis == 1 ) s = nf( v, digis, 2 );
  else if ( digis == 2 ) s = nf( v, digis, 1 ); 
  else  s = nf( (int)v, digis );

  return s;
}

String numToStringOneDecimal( float v )
{
  int digis=5;
  String s;
  if ( abs(v) < 10 ) digis = 1;
  else if ( abs(v) < 100 ) digis = 2;
  else if ( abs(v) < 1000 ) digis = 3;
  else if ( abs(v) < 10000 ) digis = 4;
  s = nf( v, digis, 1 );

  return s;
}





