class GroupElement
{
  String _embryo_id;
  int _cell_id; // for now we are assuming numerical id's
  float _value;
  float _normalized_value;
  boolean _render, _flagged_for_removal;
  boolean _from_an_existing_group;
  int _existing_group_index;
  
  GroupElement( String e_id, int c_id )
  {
    _embryo_id = e_id;
    _cell_id = c_id;
    _value = 0.0;
    _render = _flagged_for_removal = false;
    _from_an_existing_group = false;
    _existing_group_index = -1;
  }
  
  GroupElement( String e_id, int c_id, float v )
  {
    _embryo_id = e_id;
    _cell_id = c_id;
    _value = v;
    _normalized_value = v;
    _render = _flagged_for_removal = false;
    _from_an_existing_group = false;
    _existing_group_index = -1;
  }
  
  GroupElement( GroupElement other )
  {
    _embryo_id = other._embryo_id;
    _cell_id = other._cell_id;
    _value = other._value;
    _normalized_value = other._value;
    _render = other._render;
    _flagged_for_removal = other._flagged_for_removal;
    _from_an_existing_group = other._from_an_existing_group;
    _existing_group_index = other._existing_group_index;
  }
  
  int getCellID()
  {
    return _cell_id;
  }
  
  float getValue()
  {
    return _value;
  }
  
  String getEmbryoID()
  {
    return _embryo_id;
  }
  
  void normalizeValue( float min_v, float max_v )
  {
    _normalized_value = (_value - min_v) / (max_v - min_v);
  }
  
  float getNormalizedValue()
  {
    return _normalized_value;
  }
  
  void render( boolean r )
  {
    _render = r;
  }
  
  boolean render()
  {
    return _render;
  }
  
  void flagForRemoval( boolean f )
  {
    _flagged_for_removal = f;
  }
  
  boolean flaggedForRemoval()
  {
    return _flagged_for_removal;
  }
  
  void fromAnExistingGroup( boolean f )
  {
    _from_an_existing_group = f;
  }
  
  void existingGroupIndex( int e )
  {
    _existing_group_index = e;
  }
  
  boolean fromAnExistingGroup()
  {
    return _from_an_existing_group;
  }
  
  int existingGroupIndex()
  {
    return _existing_group_index;
  }
}

/************************************************************************/

class Group
{
  String _id;
  ArrayList _elements;
  boolean _values_normalized;
  float _min_value, _max_value;
  color _color;
  
  Group( String id )
  {
    _id = id;
    _elements = new ArrayList();
    _values_normalized = false;
    _min_value = Float.MAX_VALUE;
    _max_value = Float.MIN_VALUE;
    _color = color(0);
  }
  
  Group( Group other )
  {
    _id = other._id;
    _elements = new ArrayList();
    for ( int i = 0; i < other._elements.size(); i++ )
      addElement( new GroupElement(other.getElement(i)) );
    _values_normalized = other._values_normalized;
    _min_value = other._min_value;
    _max_value = other._max_value;
    _color = other._color;
  }
  
  void id( String i )
  {
    _id = i;
  }
  
  String getID()
  {
    return _id;
  }
  
  void setColor( color c )
  {
    _color = c;
  }
  
  color getColor()
  {
    return _color;
  }
  
  void addElement( GroupElement e )
  {
    _elements.add( e );
    //_min_value = min( _min_value, e.getValue() );
    //_max_value = max( _max_value, e.getValue() );
  }
  
  GroupElement getElement( int index )
  {
    return (GroupElement)_elements.get( index );
  }
  
  void removeElement( int index )
  {
    _elements.remove( index );
  }
  
  int getNumElements()
  {
    return _elements.size();
  }
  
  void clearGroup()
  {
    Embryo e;
    GroupElement ge;
    for ( int i = 0; i < _elements.size(); i++ )
    {
      ge = (GroupElement)_elements.get(i);
      e = _embryos[embryoID2Index(ge.getEmbryoID())];
      e.getCell(e.cellID2Index(ge.getCellID())).inCreatedGroup(false);
    }
    _elements.clear();
  }
  
  boolean valuesNormalized()
  {
    return _values_normalized;
  }
  
  void normalizeValues()
  {
    _min_value = Float.MAX_VALUE;
    _max_value = Float.MIN_VALUE;
   
    // find the min and max values in the group
    for ( int i = 0; i < getNumElements(); i++ )
    {
      _min_value = min( _min_value, getElement(i).getValue() );
      _max_value = max( _max_value, getElement(i).getValue() );
    }
    
    // normalize the values in the group
    for ( int i = 0; i < _elements.size(); i++ )
      getElement(i).normalizeValue( _min_value, _max_value );
  }
  
  float getMinValue()
  {
    return _min_value;
  }
  
  float getMaxValue()
  {
    return _max_value;
  }
}











