import processing.pdf.*;

import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;

int _w, _h, _min_w, _min_h;

EmbryoView _embryo_view;
int[] _embryo_view_origin;
int _embryo_view_width, _embryo_view_height;

AggregationView _aggregation_view;
int[] _aggregation_view_origin;
int _aggregation_view_width, _aggregation_view_height;

CurvemapView _curvemap_view;
int[] _curvemap_view_origin;
int _curvemap_view_width, _curvemap_view_height;

int _divider_pos, _divider_pos_max;
float _left_percentage, _right_percentage;
boolean _divider_line_selected = false;
boolean _divider_line_hover = false;
int _diff_divider_curvemap;

boolean _right_pressed = false;

PImage _move_img, _other_move_img;
int _move_img_w, _move_img_h;

boolean _print_pdf = false;
int _pdf_counter = 0;

void setup() 
{ 
  // to capture the size of the monitor
  _w = screen.width;
  _h = screen.height-40;

  // but lets just hardcode this for now
  _w = min( _w, 1200);
  _h = min( _h, 800);

  // min width and heigh
  _min_w = 350;
  _min_h = 500;

  size( _w, _h, P2D );

  println( "Generating fonts" );
  _pixel_font_8 = createFont( "PFTempestaSeven", 8, false );
  _pixel_font_8b = createFont( "PFTempestaSeven-Bold", 8, false );
  _verdana_font_16 = createFont( "Verdana", 20, true );
  _verdana_font_30 = createFont( "Verdana", 30, true );
  textMode( SCREEN );

  // generate the data views
  _embryo_view = new EmbryoView();
  _embryo_view_origin = new int[2];

  _aggregation_view = new AggregationView();
  _aggregation_view_origin = new int[2];

  _curvemap_view = new CurvemapView();
  _curvemap_view_origin = new int[2];

  // determine the initial location of the divider
  _divider_pos_max = 3*WINDOW_BORDER_WIDTH/2 + _embryo_view.getMaxWidth();
  _divider_pos = min(min(_divider_pos_max,_w-_min_w), _w);
  _diff_divider_curvemap = WINDOW_BORDER_WIDTH/2;
  setPercentages();

  // and set the dimensions
  setDimensions();

  // set up for window resizing
  frame.setResizable(true);
  frame.addComponentListener(new ComponentAdapter() {
    public void componentResized(ComponentEvent e) {
      if(e.getSource()==frame) 
      { 
        _w = max( frame.getWidth(), _min_w*2 );
        _h = max( frame.getHeight(), _min_h+22 );
        frame.setSize(_w,_h); 
        _h -= 22; // offset on mac
        _divider_pos = max(min(/*(int)(_left_percentage*(float)_w)*/_divider_pos, min(_divider_pos_max,_w-_min_w)), _min_w);
        setPercentages();
        setDimensions();
        //_curvemap_view.initializeCompGroup();
      }
    }
  } 
  );

  // read in the data
  readConfigFile( "config.txt" );
  //readConfigFile( "config_zeba.txt" );
  
  // do some checks
  checkExistanceOfAggregations();
  initializations();
  
  makeMoveImage();
}

void setPercentages()
{
  _left_percentage = (float)_divider_pos/(float)_w;
  _right_percentage = 1.0 - _left_percentage;
}

void setDimensions()
{ 
  // set-up the AGGREGATION VIEW
  _aggregation_view_origin[0] = 2*WINDOW_BORDER_WIDTH;
  _aggregation_view_origin[1] = 3*WINDOW_BORDER_WIDTH/2;
  _aggregation_view_width = (int)(_left_percentage*(float)_w) - 7*WINDOW_BORDER_WIDTH/2;
  _aggregation_view_height = AGGREGATION_VIEW_HEIGHT;
  _aggregation_view.setDimensions( _aggregation_view_width, _aggregation_view_height );


  // set-up the EMBRYO VIEW
  _embryo_view_origin[0] = WINDOW_BORDER_WIDTH;
  _embryo_view_origin[1] = 3*WINDOW_BORDER_WIDTH/2 + _aggregation_view_height + WINDOW_BORDER_WIDTH;
  _embryo_view.setDimensions( (int)(_left_percentage*(float)_w) - 3*WINDOW_BORDER_WIDTH/2, (_h - _embryo_view_origin[1] - WINDOW_BORDER_WIDTH) );
  _embryo_view_width = _embryo_view.getViewWidth();
  _embryo_view_height = _embryo_view.getViewHeight();

  // set-up the CURVEMAP VIEW
  _curvemap_view_origin[0] = (int)((1.0-_right_percentage)*(float)_w) + WINDOW_BORDER_WIDTH/2;
  _curvemap_view_origin[1] = WINDOW_BORDER_WIDTH;
  _curvemap_view_width = (int)(_right_percentage*(float)_w) - 3*WINDOW_BORDER_WIDTH/2;
  _curvemap_view_height = _h - 2*WINDOW_BORDER_WIDTH;
  _curvemap_view.setDimensions( _curvemap_view_width, _curvemap_view_height );
  
}

void makeMoveImage()
{
  _move_img_w = 15;
  _move_img_h = 11;
  _move_img = createImage( _move_img_w, _move_img_h, ARGB );
  for ( int i = 0; i < _move_img_w*_move_img_h; i++ )
    _move_img.pixels[i] = color( 255, 0 );
  
  int i, j;
  i = 0;
  j = 5;
  _move_img.pixels[i+j*_move_img_w] = MOVE_IMAGE_COLOR;
  _move_img.pixels[(_move_img_w-1-i)+j*_move_img_w] = MOVE_IMAGE_COLOR;
  
  i = 1;
  for ( j = 4; j <=6; j++ )
  {
    _move_img.pixels[i+j*_move_img_w] = MOVE_IMAGE_COLOR;
    _move_img.pixels[(_move_img_w-1-i)+j*_move_img_w] = MOVE_IMAGE_COLOR;
  }
  
  i = 2;
  for ( j = 3; j <=7; j++ )
  {
    _move_img.pixels[i+j*_move_img_w] = MOVE_IMAGE_COLOR;
    _move_img.pixels[(_move_img_w-1-i)+j*_move_img_w] = MOVE_IMAGE_COLOR;
  }
  
  i = 3;
  for ( j = 2; j <=8; j++ )
  {
    _move_img.pixels[i+j*_move_img_w] = MOVE_IMAGE_COLOR;
    _move_img.pixels[(_move_img_w-1-i)+j*_move_img_w] = MOVE_IMAGE_COLOR;
  }
  
  i = 4;
  for ( j = 1; j <=9; j++ )
  {
    _move_img.pixels[i+j*_move_img_w] = MOVE_IMAGE_COLOR;
    _move_img.pixels[(_move_img_w-1-i)+j*_move_img_w] = MOVE_IMAGE_COLOR;
  }
  
  
  i = 7;
  for ( j = 0; j <_move_img_h; j++ ) _move_img.pixels[i+j*_move_img_w] = MOVE_IMAGE_COLOR;
  
  _other_move_img = createImage( _move_img_h, _move_img_w, ARGB );
  for ( i = 0; i < _move_img_h; i++ )
    for ( j = 0; j < _move_img_w; j++ )
    _other_move_img.pixels[i+j*_move_img_h] = _move_img.pixels[j+i*_move_img_w];
}


void draw() 
{ 
  // this is for rendering to a pdf file
  if ( _print_pdf ) {
    String filename = "output_" + nf(_pdf_counter++,2) + ".pdf";
    beginRecord(PDF, filename);
    _use_text_offset = false;
  }

  background( BACKGROUND_COLOR ); 
  
//  textFont( _verdana_font_30 );
//  fill( KEY_COLOR );
//  textAlign( LEFT, TOP );
//  text( "ExpressionExplorer", 10, 10 );

  // translate to the origin of the AGGREGATION VIEW
  pushMatrix();
  translate( _aggregation_view_origin[0], _aggregation_view_origin[1] );
  _aggregation_view.render();
  popMatrix();

  // translate to the origin of the EMBRYO VIEW
  pushMatrix();
  translate( _embryo_view_origin[0], _embryo_view_origin[1] );
  _embryo_view.render();
  popMatrix();

  // translate to the origin of the CURVEMAP VIEW
  pushMatrix();
  translate( _curvemap_view_origin[0], _curvemap_view_origin[1] );
  _curvemap_view.render();
  popMatrix();

  // draw the divider line
  if ( _divider_line_selected || _divider_line_hover )
  {
    noCursor();
    image( _move_img, mouseX-_move_img_w/2, mouseY-_move_img_h/2 );
  }
  else if ( _curvemap_divider_line_selected || _curvemap_divider_line_hoover )
  {
    noCursor();
    image( _other_move_img, mouseX-_move_img_h/2, mouseY-_move_img_w/2 );
  }
  else if ( _embryo_map_hover || _comp_group_map_hover ) cursor( CROSS );
  else if ( _print_pdf ) noCursor();
  else cursor( ARROW );

  // for turning off rendering to a pdf file
  if ( _print_pdf ) {
    _print_pdf = false;
    endRecord();
    _use_text_offset = true;
  }
} 


void mousePressed() 
{
  boolean r_value;
  if ( (mouseX >= _aggregation_view_origin[0]) && (mouseX <= _aggregation_view_origin[0]+_aggregation_view_width) &&
    (mouseY >= _aggregation_view_origin[1]) && (mouseY <= _aggregation_view_origin[1]+_aggregation_view_height+10) )
  {
    r_value = _aggregation_view.mousePressedInView( mouseX-_aggregation_view_origin[0], mouseY-_aggregation_view_origin[1] );
    if ( r_value )
    {
      _embryo_view.setSliderValues();
      _curvemap_view.initializeCompGroup();
      _curvemap_view.setScrollDimensions();
    }
    _curvemap_view.mousePressedSomewhere();
  }

  else if ( (mouseX >= _embryo_view_origin[0]) && (mouseX <= _embryo_view_origin[0]+_embryo_view_width) &&
    (mouseY >= _embryo_view_origin[1]) && (mouseY <= _embryo_view_origin[1]+_embryo_view_height) )
  {
    _right_pressed = (mouseButton == RIGHT );
    _embryo_view.mousePressedInView( mouseX-_embryo_view_origin[0], mouseY-_embryo_view_origin[1], _right_pressed );
    if ( !_right_pressed ) _curvemap_view.initializeCompGroup();
    _curvemap_view.setScrollDimensions();
    
    _curvemap_view.mousePressedSomewhere();
  }
  else if ( ((mouseX-_diff_divider_curvemap) >= _divider_pos-DIVIDER_LINE_WIDTH) && 
    ((mouseX-_diff_divider_curvemap) <= _divider_pos+DIVIDER_LINE_WIDTH) )
    {
    _divider_line_selected = true;
    _curvemap_view.mousePressedSomewhere();
    }

  else if ( (mouseX >= _curvemap_view_origin[0]) && (mouseX <= _curvemap_view_origin[0]+_curvemap_view_width) &&
    (mouseY >= _curvemap_view_origin[1]) && (mouseY <= _curvemap_view_origin[1]+_curvemap_view_height) )
    {
      _curvemap_view.mousePressedInView( mouseX-_curvemap_view_origin[0], mouseY-_curvemap_view_origin[1], (mouseButton == RIGHT) );
    }

}

void mouseReleased() 
{
  _embryo_view.mouseReleasedSomewhere();
  _aggregation_view.mouseReleasedSomewhere();
  _curvemap_view.mouseReleasedSomewhere();
  _divider_line_selected = false;
  _divider_line_hover = false;
  _curvemap_divider_line_selected = false;
  _curvemap_divider_line_hoover = false;
  _right_pressed = false;
}

void mouseMoved() 
{
  _divider_line_hover = false; 
  _embryo_map_hover = false;
  _comp_group_map_hover = false;
  _curvemap_divider_line_hoover = false;
  _embryo_view.cellRolledOverReinit();
  if ( ((mouseX-_diff_divider_curvemap) >= _divider_pos-DIVIDER_LINE_WIDTH) && 
    ((mouseX-_diff_divider_curvemap) <= _divider_pos+DIVIDER_LINE_WIDTH) )
    _divider_line_hover = true;
  else if ( (mouseX >= _embryo_view_origin[0]) && (mouseX <= _embryo_view_origin[0]+_embryo_view_width) &&
    (mouseY >= _embryo_view_origin[1]) && (mouseY <= _embryo_view_origin[1]+_embryo_view_height) )
  {
    _embryo_map_hover = _embryo_view.mouseMovedInView( mouseX-_embryo_view_origin[0], mouseY-_embryo_view_origin[1] );
  }
  else if ( (mouseX >= _curvemap_view_origin[0]) && (mouseX <= _curvemap_view_origin[0]+_curvemap_view_width) &&
    (mouseY >= _curvemap_view_origin[1]) && (mouseY <= _curvemap_view_origin[1]+_curvemap_view_height) )
    _curvemap_view.mouseMovedInView( mouseX-_curvemap_view_origin[0], mouseY-_curvemap_view_origin[1] );
}

void mouseDragged() 
{
  if ( _divider_line_selected )
  {
    _divider_pos = max(min( (mouseX-_diff_divider_curvemap), min(_divider_pos_max,_w-_min_w) ), _min_w);
    setPercentages(); 
    setDimensions();
  }
  else if ( (mouseX >= 0) && (mouseX <= _embryo_view_origin[0]+_embryo_view_width) &&
    (mouseY >= _embryo_view_origin[1]) && (mouseY <= _embryo_view_origin[1]+_embryo_view_height) )
  {
    _embryo_view.mouseDraggedInView( mouseX-_embryo_view_origin[0], mouseY-_embryo_view_origin[1], _right_pressed );
    if ( mouseButton == RIGHT ) _curvemap_view.setScrollDimensions();
  }
  else if ( (mouseX >= _aggregation_view_origin[0]) && (mouseX <= _aggregation_view_origin[0]+_aggregation_view_width) &&
    (mouseY >= _aggregation_view_origin[1]) && (mouseY <= _aggregation_view_origin[1]+_aggregation_view_height) )
    _aggregation_view.mouseDraggedInView( mouseX-_aggregation_view_origin[0], mouseY-_aggregation_view_origin[1] );
  else if ( (mouseX >= _curvemap_view_origin[0]) && (mouseX <= _curvemap_view_origin[0]+_curvemap_view_width) &&
    (mouseY >= _curvemap_view_origin[1]) && (mouseY <= _curvemap_view_origin[1]+_curvemap_view_height) )
    _curvemap_view.mouseDraggedInView( mouseX-_curvemap_view_origin[0], mouseY-_curvemap_view_origin[1], (mouseButton == RIGHT) );
}

void keyReleased() 
{
}

void keyPressed() 
{
  if ( _curvemap_view.keyPressedInView() )
  {
  }
  else if ( key == 'P' ) _print_pdf = true;
}





