class SaveLoadView
{
  int _width, _height;
  int[] _text_offset, _save_pos, _load_pos, _loaded_groups_pos;
  int _text_box_width, _text_box_height, _button_width, _button_height, _save_load_offset, _label_offset;
  int _loaded_groups_width, _loaded_groups_max_width, _loaded_groups_offset;
  boolean _save_selected, _load_selected, _load_index_selected;

  int[] _load_triangle_pos;
  int _load_triangle_height, _load_line_height;

  int[] _save_button_pos, _load_button_pos;
  String _save_name;

  int _load_selected_index;
  ArrayList _loaded_groups;
  boolean _loaded_group_rolled_over;
  int _rolled_over_loaded_group;

  SaveLoadView()
  {
    _text_offset = new int[2];
    _save_pos = new int[2];
    _load_pos = new int[2];
    _load_triangle_pos = new int[2];
    _save_button_pos = new int[2];
    _load_button_pos = new int[2];
    _loaded_groups_pos = new int[2];

    _text_box_height = SAVE_LOAD_VIEW_TEXT_BOX_HEIGHT;
    _button_width = SAVE_LOAD_VIEW_BUTTON_WIDTH;
    _button_height = SAVE_LOAD_VIEW_BUTTON_HEIGHT;
    _save_load_offset = SAVE_LOAD_VIEW_OFFSET;
    _label_offset = SAVE_LOAD_VIEW_LABEL_OFFSET;
    _load_triangle_height = SAVE_LOAD_VIEW_TRIANGLE_HEIGHT;
    _load_line_height = SAVE_LOAD_VIEW_LOAD_LINE_HEIGHT;

    _save_selected = _load_selected = _load_index_selected = false;

    _save_name = "";
    _load_selected_index = -1;
    
    _loaded_group_rolled_over = false;
    _rolled_over_loaded_group = 0;
    _loaded_groups_offset = SAVE_LOAD_VIEW_LOADED_GROUPS_OFFSET;
    _loaded_groups = new ArrayList();
    setLoadedGroupsWidth();
  }
  
  boolean loadedGroupRolledOver()
  {
    return _loaded_group_rolled_over;
  }
  
  int rolledOverLoadedGroup()
  {
    return _rolled_over_loaded_group;
  }
  
  boolean loadSelected()
  {
    return _load_selected;
  }

  void setDimensions( int w, int h )
  {
    _width = w;
    _height = h;

    _text_box_width = min( (_width-_save_load_offset)/2, SAVE_LOAD_VIEW_TEXT_BOX_MAX_WIDTH );

    _save_pos[0] = 0;
    _save_pos[1] = _label_offset;
    _load_pos[0] = (_width+_save_load_offset)/2;
    _load_pos[1] = _label_offset;    

    _save_button_pos[0] = _save_pos[0] + _text_box_width - _button_width;
    _save_button_pos[1] = _save_pos[1] + _text_box_height + 5;

    _load_button_pos[0] = _load_pos[0] + _text_box_width - _button_width;
    _load_button_pos[1] = _load_pos[1] + _text_box_height + 5;

    _load_triangle_pos[0] = _load_pos[0] + 97;
    _load_triangle_pos[1] = _load_pos[1] - _load_triangle_height - 2;
    
    _loaded_groups_pos[0] = 0;
    _loaded_groups_pos[1] = _save_button_pos[1] + _button_height + 10;
    _loaded_groups_max_width = _width;
  }
  
  void setLoadedGroupsWidth()
  {
    // fill the laoded groups list
    GroupElement ge;
    int g_index;
    _loaded_groups.clear();
    for ( int i = 0; i < _created_group.getNumElements(); i++ )
    {
      ge = _created_group.getElement(i);
      if ( !ge.fromAnExistingGroup() ) continue;
      
      g_index = ge.existingGroupIndex();
      boolean exists = false;
      for ( int j = 0; j < _loaded_groups.size(); j++ )
      {
        if ( g_index == (Integer)_loaded_groups.get(j) )
        {
          exists = true;
          break;
        }
      }
        
      if ( !exists )
        _loaded_groups.add( g_index );
    }
    
    _loaded_groups_width = 0;
    for ( int i = 0; i < _loaded_groups.size(); i++ )
      _loaded_groups_width += textWidth( ((Group)_existing_groups.get((Integer)_loaded_groups.get(i))).getID() ) + _loaded_groups_offset;
    _loaded_groups_width -= _loaded_groups_offset;
  }

  void setTextOffset( int tx, int ty )
  {
    _text_offset = new int[2];
    _text_offset[0] = tx;
    _text_offset[1] = ty;
  }

  void render()
  {
    int x, y, tx_offset=0, ty_offset=0;

    //
    // the save stuff
    //
    if( _use_text_offset )
    {
      tx_offset = _text_offset[0];
      ty_offset = _text_offset[1];
    }
    int y_bottom_offset = 0;
    if ( !_use_text_offset )
    { 
      ty_offset = -PDF_OFFSET;
      y_bottom_offset = PDF_OFFSET-1;
    }

    x = _save_pos[0];
    y = _save_pos[1];
    stroke( SAVE_LOAD_VIEW_TEXT_BOX_OUTLINE_COLOR );
    strokeWeight( SAVE_LOAD_VIEW_OUTLINE_WEIGHT );

    if ( _save_selected ) fill( SAVE_LOAD_VIEW_TEXT_BOX_SELECTED_FILL_COLOR );
    else fill( SAVE_LOAD_VIEW_TEXT_BOX_FILL_COLOR );

    rect( x, y, _text_box_width, _text_box_height );

    textFont(_pixel_font_8); 
    textAlign( LEFT, BOTTOM );
    fill( SAVE_LOAD_VIEW_TEXT_BOX_TEXT_COLOR );
    text( "Save Created Group", x+tx_offset, y-2+ty_offset+y_bottom_offset );

    textAlign( LEFT, CENTER );
    if ( !_save_selected )
    {    
      text( "enter name", x+2+tx_offset, y+_text_box_height/2+ty_offset );
    }
    else
    {
      text( _save_name, x+2+tx_offset, y+_text_box_height/2+ty_offset );
    }

    // the save button
    x = _save_button_pos[0];
    y = _save_button_pos[1];
    fill( SAVE_LOAD_VIEW_BUTTON_FILL_COLOR );
    stroke( SAVE_LOAD_VIEW_BUTTON_OUTLINE_COLOR );
    rect( x, y, _button_width, _button_height );

    x += _button_width/2;
    y += _button_height/2;
    if ( _save_selected ) fill( BACKGROUND_COLOR );
    else fill( SAVE_LOAD_VIEW_BUTTON_OUTLINE_COLOR );
    textAlign( CENTER, CENTER );
    text( "save", x+tx_offset, y+ty_offset );

    //
    // the load stuff
    //
    if( _use_text_offset )
    {
      tx_offset = _text_offset[0];
      ty_offset = _text_offset[1];
    }

    x = _load_pos[0];
    y = _load_pos[1];

    // the label
    textFont(_pixel_font_8); 
    textAlign( LEFT, BOTTOM );
    fill( SAVE_LOAD_VIEW_TEXT_BOX_TEXT_COLOR );
    text( "Load Existing Group", x+tx_offset, y-2+ty_offset+y_bottom_offset );

    if ( !_load_selected )
    {
      stroke( SAVE_LOAD_VIEW_TEXT_BOX_OUTLINE_COLOR );
      strokeWeight( SAVE_LOAD_VIEW_OUTLINE_WEIGHT );
      if ( !_load_index_selected ) fill( SAVE_LOAD_VIEW_TEXT_BOX_FILL_COLOR );
      else fill( SAVE_LOAD_VIEW_TEXT_BOX_SELECTED_FILL_COLOR );
      rect( x, y, _text_box_width, _text_box_height );

      if ( _load_index_selected )
      {
        fill( SAVE_LOAD_VIEW_TEXT_BOX_TEXT_COLOR );
        textAlign( LEFT, CENTER );
        text( ((Group)_existing_groups.get(_load_selected_index)).getID(), x+2+tx_offset, y+_text_box_height/2+ty_offset );
      }

      // the triangle
      fill( SAVE_LOAD_VIEW_TRIANGLE_FILL_COLOR );
      stroke( SAVE_LOAD_VIEW_TRIANGLE_OUTLINE_COLOR );
      x = _load_triangle_pos[0];
      y = _load_triangle_pos[1];
      triangle( x, y, x+_load_triangle_height/2, y+_load_triangle_height-2, x+_load_triangle_height, y );

      // the load button
      x = _load_button_pos[0];
      y = _load_button_pos[1];
      fill( SAVE_LOAD_VIEW_BUTTON_FILL_COLOR );
      stroke( SAVE_LOAD_VIEW_BUTTON_OUTLINE_COLOR );
      rect( x, y, _button_width, _button_height );

      x += _button_width/2;
      y += _button_height/2;
      if ( _load_index_selected ) fill( BACKGROUND_COLOR );
      else fill( SAVE_LOAD_VIEW_BUTTON_OUTLINE_COLOR );
      textAlign( CENTER, CENTER );
      text( "load", x+tx_offset, y+ty_offset );
    }
    else
    {
      stroke( SAVE_LOAD_VIEW_TEXT_BOX_OUTLINE_COLOR );
      strokeWeight( SAVE_LOAD_VIEW_OUTLINE_WEIGHT );
      fill( SAVE_LOAD_VIEW_TEXT_BOX_SELECTED_FILL_COLOR );
      int h = max( _existing_groups.size(), 1 ) * _load_line_height + _text_box_height;
      rect( x, y, _text_box_width, h );

      for ( int i = 0; i < _existing_groups.size(); i++ )
      {
        if ( _load_selected && (_load_selected_index == i) ) textFont( _pixel_font_8b );
        else textFont(_pixel_font_8); 
        textAlign( LEFT, BOTTOM );
        fill( SAVE_LOAD_VIEW_TEXT_BOX_TEXT_COLOR );
        text( ((Group)_existing_groups.get(i)).getID(), x+tx_offset, y+ty_offset+(i+1)*_load_line_height+_text_box_height+y_bottom_offset );
      }

      // the triangle
      fill( SAVE_LOAD_VIEW_TRIANGLE_SELECTED_FILL_COLOR );
      stroke( SAVE_LOAD_VIEW_TRIANGLE_OUTLINE_COLOR );
      x = _load_triangle_pos[0];
      y = _load_triangle_pos[1];
      triangle( x, y, x+_load_triangle_height/2, y+_load_triangle_height-2, x+_load_triangle_height, y );
    }
    
    if ( _loaded_groups_width > 0 )
    {
      x = _loaded_groups_pos[0];
      y = _loaded_groups_pos[1];
      
      textFont(_pixel_font_8); 
      textAlign( LEFT, BOTTOM );
      fill( SAVE_LOAD_VIEW_TEXT_BOX_TEXT_COLOR );
      text( "Loaded Groups", tx_offset+x, ty_offset+y+y_bottom_offset );
      
      textAlign( LEFT, TOP );
      Group g;
      y += 4;
      for ( int i = 0; i < _loaded_groups.size(); i++ )
      {
        g = (Group)_existing_groups.get((Integer)_loaded_groups.get(i));
        fill( g.getColor() );
        if ( _loaded_group_rolled_over && (_rolled_over_loaded_group == (Integer)_loaded_groups.get(i)) )
          textFont(_pixel_font_8b);
        else 
          textFont(_pixel_font_8); 
        text( g.getID(), tx_offset+x, ty_offset+y );
        
        textFont(_pixel_font_8); 
        x += (textWidth(g.getID()) + _loaded_groups_offset);
      }
    }
  }

  void mouseReleasedSomewhere()
  {
    if ( _load_selected )
    {
      if ( _load_selected_index != -1 ) _load_index_selected = true;
      _load_selected = false;
    }
  }

  void mousePressedSomewhere()
  {
    _save_selected = false;
    _save_name = "";
  }

  boolean mousePressedInView( int mx, int my )
  {
    if ( (mx >= _save_pos[0]) && (mx <= _save_pos[0]+_text_box_width) && (my >= _save_pos[1]) && (my <= _save_pos[1]+_text_box_height) )
    {
      _save_selected = true;
      //_load_selected = false;
    }

    else if ( _save_selected && overSaveButton(mx, my) )
    {
      if ( _save_name.length() != 0 )
        saveCreatedGroup();
    }

    else if ( _load_index_selected && overLoadButton(mx, my) )
    {
      loadGroup();
    }

    else if ( (mx >= _load_pos[0]) && (mx <= _load_pos[0]+_text_box_width) && (my >= _load_pos[1]) && (my <= _load_pos[1]+_text_box_height) )
    {
      _load_selected = true;
      _load_selected_index = -1;
      mousePressedSomewhere();
    }

    else if ( (mx >= _load_triangle_pos[0]) && (mx <= _load_triangle_pos[0]+_load_triangle_height) && 
      (my >= _load_triangle_pos[1]) && (my <= _load_triangle_pos[1]+_load_triangle_height) )
    {
      _load_selected = true;
      mousePressedSomewhere();
    }
    return false;
  }

  void mouseDraggedInView( int mx, int my )
  {
    if ( _load_selected )
    {
      _load_selected_index = -1;

      // check for over a name
      int x = _load_pos[0];
      int y = _load_pos[1]+_text_box_height;

      if ( (mx < x) || (mx > x+_text_box_width) ) return;

      for ( int i = 0; i < _existing_groups.size(); i++ )
      {
        if ( (my >= y) && (my <= y+_load_line_height) ) _load_selected_index = i;
        y += _load_line_height;
      }
    }
  }
  
  void mouseMovedInView( int mx, int my )
  {
    _loaded_group_rolled_over = false;
    if ( (my >= _loaded_groups_pos[1]) && (my <= _loaded_groups_pos[1]+10) )
    {
      int x = 0;
      int w;
      for ( int i = 0; i < _loaded_groups.size(); i++ )
      {
        w = (int)textWidth( ((Group)_existing_groups.get((Integer)_loaded_groups.get(i))).getID() );
        if ( (mx >= x) && (mx <= x+w) )
        {
          _loaded_group_rolled_over = true;
          _rolled_over_loaded_group = (Integer)_loaded_groups.get(i);
        }
        x += w+_loaded_groups_offset;
      }
    }
  }

  boolean keyPressedInView()
  {
    if ( _save_selected )
    {
      if ( ((key >= 48) && (key <= 57 )) || ((key >= 65) && (key <= 90)) || ((key >= 97) && (key <= 122)) || (key == 45) || (key == 95) ) 
      {
        _save_name += Character.toString(key);
      }

      else if (key == BACKSPACE ) 
      {
        _save_name = _save_name.substring(0, _save_name.length()-1 );
      }
      else if ( (key == RETURN) || (key == ENTER) ) 
      {
        saveCreatedGroup();
      }

      return true;
    }
    else
    {
      if ( (key == RETURN) || (key == ENTER) ) loadGroup();
    }
    return false;
  }

  boolean overSaveButton( int mx, int my )
  {
    if ( (mx >= _save_button_pos[0]) && (mx <= _save_button_pos[0]+_button_width) && (my >= _save_button_pos[1]) && (my <= _save_button_pos[1]+_button_height) )
      return true;
    return false;
  }

  boolean overLoadButton( int mx, int my )
  {
    if ( (mx >= _load_button_pos[0]) && (mx <= _load_button_pos[0]+_button_width) && (my >= _load_button_pos[1]) && (my <= _load_button_pos[1]+_button_height) )
      return true;
    return false;
  }

  void saveCreatedGroup()
  {
    // create the file
    int num = _created_group.getNumElements()+3;
    String[] rows = new String[num];

    rows[0] = "## created on " + month() + "/" + day() + "/" + year() + " at " + hour() + ":" + minute() + " using ExpressionExplorer";
    rows[1] = "#NAME: " + _save_name;
    rows[2] = "#species,cell_id";

    GroupElement ge;
    for ( int i = 3; i < num; i++ )
    {
      ge = _created_group.getElement(i-3);
      rows[i] = ge.getEmbryoID() + "," + ge.getCellID();
    }

    saveStrings( _save_name+".csv", rows );

    _created_group.id( _save_name );
    _existing_groups.add( new Group(_created_group) );
    
    // add the existing group flag to the elements
    Group g = (Group)_existing_groups.get(_existing_groups.size()-1);
    g.setColor( COLORMAP_CATAGORICAL[(_existing_groups.size()-1) % (COLORMAP_CATAGORICAL.length)] );
    for ( int i = 0; i < g.getNumElements(); i++ )
    {
      ge = g.getElement(i);
      ge.fromAnExistingGroup( true );
      ge.existingGroupIndex( _existing_groups.size()-1 );
    }

    //_created_group.clearGroup();
    //_curvemap_view.setScrollYDimensions();

    _save_selected = false;
    _save_name = "";
    setLoadedGroupsWidth();
  }

  void loadGroup()
  {
    Group g = (Group)_existing_groups.get(_load_selected_index);
    GroupElement ge;
    for ( int i = 0; i < g.getNumElements(); i++ )
    {
      ge = g.getElement(i);
      Embryo e = _embryos[embryoID2Index(ge.getEmbryoID())];
      Cell c = e.getCell( e.cellID2Index(ge.getCellID()) );
      c.inCreatedGroup( true );
      GroupElement oge = new GroupElement( ge );
      oge.render( true );
      _created_group.addElement( oge );
    }

    _load_index_selected = false;
    _load_selected_index = -1;
    _curvemap_view.setScrollYDimensions();
    
    setLoadedGroupsWidth();
  }

}























